# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 2022-07-17
        git sha              : :%H$
        copyright            : (C) 2022 by Dave Signer
        email                : david at opengis ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import os
from typing import Union

import yaml
from qgis.core import (
    Qgis,
    QgsLayerDefinition,
    QgsLayerTreeGroup,
    QgsLayerTreeLayer,
    QgsLayerTreeNode,
    QgsMapLayer,
    QgsProject,
)
from qgis.PyQt.QtCore import QObject, pyqtSignal

from .exportsettings import ExportSettings
from .target import Target
from .utils import slugify


class ProjectTopping(QObject):
    """
    A project configuration resulting in a YAML file that contains:
    - layertree
    - layerorder
    - project variables (future)
    - print layout (future)
    - map themes (future)
    QML style files, QLR layer definition files and the source of a layer can be linked in the YAML file and are exported to the specific folders.
    """

    stdout = pyqtSignal(str, int)

    PROJECTTOPPING_TYPE = "projecttopping"
    LAYERDEFINITION_TYPE = "layerdefinition"
    LAYERSTYLE_TYPE = "layerstyle"

    class TreeItemProperties(object):
        """
        The properties of a node (tree item)
        """

        def __init__(self):
            # if the node is a group
            self.group = False
            # if the node is visible
            self.checked = True
            # if the node is expanded
            self.expanded = True
            # if the (layer) node shows feature count
            self.featurecount = False
            # if the (group) node handles mutually-exclusive
            self.mutually_exclusive = False
            # if the (group) node handles mutually-exclusive, the index of the checked child
            self.mutually_exclusive_child = -1
            # the layers provider to create it from source
            self.provider = None
            # the layers uri to create it from source
            self.uri = None
            # the style file - if None then not requested
            self.qmlstylefile = None
            # the definition file - if None then not requested
            self.definitionfile = None

    class LayerTreeItem(object):
        """
        A tree item of the layer tree. Every item contains the properties of a layer and according the ExportSettings passed on parsing the QGIS project.
        """

        def __init__(self):
            self.items = []
            self.name = None
            self.properties = ProjectTopping.TreeItemProperties()
            self.temporary_toppingfile_dir = os.path.expanduser("~/.temp_topping_files")

        def make_item(
            self,
            project: QgsProject,
            node: Union[QgsLayerTreeLayer, QgsLayerTreeGroup],
            export_settings: ExportSettings,
        ):
            # properties for every kind of nodes
            self.name = node.name()
            self.properties.checked = node.itemVisibilityChecked()
            self.properties.expanded = node.isExpanded()

            definition_setting = export_settings.get_setting(
                ExportSettings.ToppingType.DEFINITION, node, node.name()
            )
            if definition_setting.get("export", False):
                self.properties.definitionfile = self._temporary_definitionfile(node)

            if isinstance(node, QgsLayerTreeGroup):
                # it's a group
                self.properties.group = True
                self.properties.mutually_exclusive = node.isMutuallyExclusive()

                if not definition_setting.get("export", False):
                    # only consider children, when the group is not exported as DEFINITION
                    index = 0
                    for child in node.children():
                        item = ProjectTopping.LayerTreeItem()
                        item.make_item(project, child, export_settings)
                        # set the first checked item as mutually exclusive child
                        if (
                            self.properties.mutually_exclusive
                            and self.properties.mutually_exclusive_child == -1
                        ):
                            if item.properties.checked:
                                self.properties.mutually_exclusive_child = index
                        self.items.append(item)
                        index += 1
            else:
                if isinstance(node, QgsLayerTreeLayer):
                    layer = node.layer()
                else:
                    # must be not recognized as QgsLayerTreeLayer (but QgsLayerTreeNode instead)
                    layer = self._layer_of_node(project, node)
                self.properties.featurecount = node.customProperty("showFeatureCount")
                source_setting = export_settings.get_setting(
                    ExportSettings.ToppingType.SOURCE, node, node.name()
                )
                if source_setting.get("export", False):
                    if layer.dataProvider():
                        if layer.dataProvider():
                            self.properties.provider = layer.dataProvider().name()
                        self.properties.uri = (
                            QgsProject.instance()
                            .pathResolver()
                            .writePath(layer.publicSource())
                        )
                qml_setting = export_settings.get_setting(
                    ExportSettings.ToppingType.QMLSTYLE, node, node.name()
                )
                if qml_setting.get("export", False):
                    self.properties.qmlstylefile = self._temporary_qmlstylefile(
                        layer,
                        QgsMapLayer.StyleCategory(
                            qml_setting.get(
                                "categories",
                                QgsMapLayer.StyleCategory.AllStyleCategories,
                            )
                        ),
                    )

        def _layer_of_node(
            self,
            project: QgsProject,
            node: QgsLayerTreeNode,
        ) -> QgsLayerTreeLayer:
            # workaround when layer has not been detected as QgsLayerTreeLayer.
            # See https://github.com/opengisch/QgisModelBaker/pull/514
            return project.mapLayersByName(node.name())[0]

        def _temporary_definitionfile(
            self, node: Union[QgsLayerTreeLayer, QgsLayerTreeGroup]
        ):
            filename_slug = f"{slugify(self.name)}.qlr"
            os.makedirs(self.temporary_toppingfile_dir, exist_ok=True)
            temporary_toppingfile_path = os.path.join(
                self.temporary_toppingfile_dir, filename_slug
            )
            QgsLayerDefinition.exportLayerDefinition(temporary_toppingfile_path, [node])
            return temporary_toppingfile_path

        def _temporary_qmlstylefile(
            self,
            layer: QgsMapLayer,
            categories: QgsMapLayer.StyleCategories = QgsMapLayer.StyleCategory.AllStyleCategories,
        ):
            filename_slug = f"{slugify(self.name)}.qml"
            os.makedirs(self.temporary_toppingfile_dir, exist_ok=True)
            temporary_toppingfile_path = os.path.join(
                self.temporary_toppingfile_dir, filename_slug
            )
            layer.saveNamedStyle(temporary_toppingfile_path, categories)
            return temporary_toppingfile_path

    def __init__(self):
        QObject.__init__(self)
        self.layertree = self.LayerTreeItem()
        self.layerorder = []

    def parse_project(
        self, project: QgsProject, export_settings: ExportSettings = ExportSettings()
    ):
        """
        Parses a project into the ProjectTopping structure. Means the LayerTreeNodes are loaded into the layertree variable and append the ExportSettings to each node. The CustomLayerOrder is loaded into the layerorder. The project is not keeped as member variable.

        :param QgsProject project: the project to parse.
        :param ExportSettings settings: defining if the node needs a source or style / definitionfiles.
        """
        root = project.layerTreeRoot()
        if root:
            self.layertree.make_item(project, project.layerTreeRoot(), export_settings)
            self.layerorder = (
                root.customLayerOrder() if root.hasCustomLayerOrder() else []
            )
            self.stdout.emit(
                self.tr("QGIS project parsed with export settings."), Qgis.Info
            )
        else:
            self.stdout.emit(
                self.tr("Could not parse the QGIS project..."), Qgis.Warning
            )
            return False
        return True

    def generate_files(self, target: Target) -> str:
        """
        Generates all files according to the passed Target.

        :param Target target: the target object containing the paths where to create the files and the path_resolver defining the structure of the link.
        """
        # generate projecttopping as a dict
        projecttopping_dict = self._projecttopping_dict(target)

        # write the yaml
        projecttopping_slug = f"{slugify(target.projectname)}.yaml"
        absolute_filedir_path, relative_filedir_path = target.filedir_path(
            ProjectTopping.PROJECTTOPPING_TYPE
        )
        with open(
            os.path.join(absolute_filedir_path, projecttopping_slug), "w"
        ) as projecttopping_yamlfile:
            yaml.dump(projecttopping_dict, projecttopping_yamlfile)
            self.stdout.emit(
                self.tr("Project Topping written to YAML file: {}").format(
                    projecttopping_yamlfile
                ),
                Qgis.Info,
            )
        return target.path_resolver(
            target, projecttopping_slug, ProjectTopping.PROJECTTOPPING_TYPE
        )

    def load_files(self, target: Target):
        """
        - [ ] Not yet implemented.
        """
        raise NotImplementedError

    def generate_project(self, target: Target) -> QgsProject:
        """
        - [ ] Not yet implemented.
        """
        return QgsProject()

    def _projecttopping_dict(self, target: Target):
        """
        Creates the layertree as a dict.
        Creates the layerorder as a list.
        And it generates and stores the toppingfiles according th the Target.
        """
        projecttopping_dict = {}
        projecttopping_dict["layertree"] = self._item_dict_list(
            target, self.layertree.items
        )
        projecttopping_dict["layerorder"] = [layer.name() for layer in self.layerorder]
        return projecttopping_dict

    def _item_dict_list(self, target: Target, items):
        item_dict_list = []
        print([item.name for item in items])
        for item in items:
            item_dict = self._create_item_dict(target, item)
            item_dict_list.append(item_dict)
        return item_dict_list

    def _create_item_dict(self, target: Target, item: LayerTreeItem):
        item_dict = {}
        item_properties_dict = {}

        if item.properties.group:
            item_properties_dict["group"] = True
            if item.properties.mutually_exclusive:
                item_properties_dict["mutually-exclusive"] = True
                item_properties_dict[
                    "mutually-exclusive-child"
                ] = item.properties.mutually_exclusive_child
        else:
            if item.properties.featurecount:
                item_properties_dict["featurecount"] = True
            if item.properties.qmlstylefile:
                item_properties_dict["qmlstylefile"] = target.toppingfile_link(
                    ProjectTopping.LAYERSTYLE_TYPE, item.properties.qmlstylefile
                )
            if item.properties.provider and item.properties.uri:
                item_properties_dict["provider"] = item.properties.provider
                item_properties_dict["uri"] = item.properties.uri

        item_properties_dict["checked"] = item.properties.checked
        item_properties_dict["expanded"] = item.properties.expanded

        if item.properties.definitionfile:
            item_properties_dict["definitionfile"] = target.toppingfile_link(
                ProjectTopping.LAYERDEFINITION_TYPE,
                item.properties.definitionfile,
            )

        if item.items:
            child_item_dict_list = self._item_dict_list(target, item.items)
            item_properties_dict["child-nodes"] = child_item_dict_list

        item_dict[item.name] = item_properties_dict
        return item_dict



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 2022-07-17
        git sha              : :%H$
        copyright            : (C) 2022 by Dave Signer
        email                : david at opengis ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from ..libs.toppingmaker import ExportSettings
from .ili2dbsettings import Ili2dbSettings
from .ilidata import IliData
from .iliprojecttopping import IliProjectTopping
from .ilitarget import IliTarget
from .metaconfig import MetaConfig



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 08/08/17
        git sha              : :%H$
        copyright            : (C) 2017 by OPENGIS.ch
        email                : info@opengis.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.core import (
    Qgis,
    QgsAttributeEditorContainer,
    QgsAttributeEditorField,
    QgsAttributeEditorRelation,
    QgsEditFormConfig,
)


class Form(object):
    def __init__(self):
        self.__elements = list()

    def elements(self):
        return self.__elements

    def create(self, layer, qgis_layer, project):
        edit_form_config = qgis_layer.editFormConfig()
        root_container = edit_form_config.invisibleRootContainer()
        root_container.clear()
        for element in self.__elements:
            root_container.addChildElement(element.create(root_container, qgis_layer))
        edit_form_config.setLayout(QgsEditFormConfig.TabLayout)
        # set nm-rel if referencing tables are junction table
        for relation in project.relations:
            if relation.referenced_layer == layer:
                # get other relations, that have the same referencing_layer and set the first as nm-rel
                for other_relation in project.relations:
                    if (
                        other_relation.referencing_field != relation.referencing_field
                        and other_relation.referencing_layer
                        == relation.referencing_layer
                        and relation.referencing_layer.is_nmrel
                        and not other_relation.referenced_layer.is_basket_table
                    ):
                        edit_form_config.setWidgetConfig(
                            relation.id, {"nm-rel": other_relation.id}
                        )
                        break
        return edit_form_config

    def add_element(self, element):
        self.__elements.append(element)


class FormTab(object):
    def __init__(self, name, columns=1):
        self.name = name
        self.children = list()
        self.columns = columns

    def addChild(self, child):
        self.children.append(child)

    def create(self, parent, layer):
        container = QgsAttributeEditorContainer(self.name, parent)
        container.setIsGroupBox(False)
        container.setColumnCount(self.columns)

        for child in self.children:
            container.addChildElement(child.create(container, layer))
        return container


class FormGroupBox(object):
    def __init__(self, name, columns=1):
        self.name = name
        self.children = list()
        self.columns = columns

    def addChild(self, child):
        self.children.append(child)

    def create(self, parent, layer):
        container = QgsAttributeEditorContainer(self.name)
        container.setIsGroupBox(True)
        container.setColumnCount(self.columns)

        for child in self.children:
            container.addChildElement(child, layer)
        return container


class FormFieldWidget(object):
    def __init__(self, name, field_name):
        self.name = name if name else field_name
        self.field_name = field_name

    def create(self, parent, layer):
        index = layer.fields().indexOf(self.field_name)
        widget = QgsAttributeEditorField(self.field_name, index, parent)
        return widget


class FormRelationWidget(object):
    def __init__(self, relation, nm_relation=None):
        self.relation = relation
        self.nm_relation = nm_relation

    def create(self, parent, layer):
        try:
            widget = QgsAttributeEditorRelation(self.relation.id, parent)
        except TypeError:
            # Feed deprecated API for 3.0.0 and 3.0.1
            widget = QgsAttributeEditorRelation(
                self.relation.id, self.relation.id, parent
            )
        if self.nm_relation:
            widget.setNmRelationId(self.nm_relation.id)

        target_layer = (
            self.nm_relation.referenced_layer
            if self.nm_relation
            else self.relation.referencing_layer
        )

        if Qgis.QGIS_VERSION_INT >= 31800:
            if not target_layer.geometry_column:
                widget.setRelationWidgetTypeId("linking_relation_editor")

                cardinality_max = self.relation.cardinality_max
                if self.nm_relation:
                    cardinality_max = self.nm_relation.cardinality_max

                if cardinality_max == "1":
                    configuration = widget.relationEditorConfiguration()
                    configuration["one_to_one"] = True
                    widget.setRelationEditorConfiguration(configuration)

        return widget



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 2022-07-17
        git sha              : :%H$
        copyright            : (C) 2022 by Dave Signer
        email                : david at opengis ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""


class Ili2dbSettings(dict):
    """
    Class keeping settings to be performed on ili2db.
    As well as the files like pre-/postscript and metaattributefile (TOML/INI).

    With the method parse_parameter_from_db we can fill it up according to a existing database schema (passing the specific extension of DBConnector).
    """

    def __init__(self):
        self.parameters = {}
        self.metaattr_path = None
        self.postscript_path = None
        self.prescript_path = None
        self.models = []

    def parse_parameters_from_db(self, db_connector):
        """
        Fill the parameter list according to a existing database schema (passed by the specific extension of DBConnector).
        It only considers the settings "known" by modelbaker (the parameters set by the user or by modelbaker per default).
        The parameter are only set when they are available (and not unset when they are not available).
        """

        setting_records = db_connector.get_ili2db_settings()
        settings_dict = {}
        self.parameters = {}
        for setting_record in setting_records:
            settings_dict[setting_record["tag"]] = setting_record["setting"]

        # user settings
        if settings_dict.get("ch.ehi.ili2db.inheritanceTrafo", None) == "smart1":
            self.parameters["smart1Inheritance"] = True
        if settings_dict.get("ch.ehi.ili2db.inheritanceTrafo", None) == "smart2":
            self.parameters["smart2Inheritance"] = True
        if settings_dict.get("ch.ehi.ili2db.StrokeArcs", None) == "enable":
            self.parameters["strokeArcs"] = True
        if settings_dict.get("ch.ehi.ili2db.BasketHandling", None) == "readWrite":
            self.parameters["createBasketCol"] = True
        if settings_dict.get("ch.ehi.ili2db.defaultSrsAuthority", None):
            self.parameters["defaultSrsAuth"] = settings_dict[
                "ch.ehi.ili2db.defaultSrsAuthority"
            ]
        if settings_dict.get("ch.ehi.ili2db.StrokedefaultSrsCodeArcs", None):
            self.parameters["defaultSrsCode"] = settings_dict[
                "ch.ehi.ili2db.defaultSrsCode"
            ]

        # modelbaker default settings
        if settings_dict.get("ch.ehi.ili2db.catalogueRefTrafo", None) == "coalesce":
            self.parameters["coalesceCatalogueRef"] = True
        if (
            settings_dict.get("ch.ehi.ili2db.createEnumDefs", None)
            == "multiTableWithId"
        ):
            self.parameters["createEnumTabsWithId"] = True
        if settings_dict.get("ch.ehi.ili2db.numericCheckConstraints", None) == "create":
            self.parameters["createNumChecks"] = True
        if settings_dict.get("ch.ehi.ili2db.uniqueConstraints", None) == "create":
            self.parameters["createUnique"] = True
        if settings_dict.get("ch.ehi.ili2db.createForeignKey", None) == "yes":
            self.parameters["createFk"] = True
        if settings_dict.get("ch.ehi.ili2db.createForeignKeyIndex", None) == "yes":
            self.parameters["createFkIdx"] = True
        if settings_dict.get("ch.ehi.ili2db.multiSurfaceTrafo", None) == "coalesce":
            self.parameters["coalesceMultiSurface"] = True
        if settings_dict.get("ch.ehi.ili2db.multiLineTrafo", None) == "coalesce":
            self.parameters["coalesceMultiLine"] = True
        if settings_dict.get("ch.ehi.ili2db.multiPointTrafo", None) == "coalesce":
            self.parameters["coalesceMultiPoint"] = True
        if settings_dict.get("ch.ehi.ili2db.arrayTrafo", None) == "coalesce":
            self.parameters["coalesceArray"] = True
        if (
            settings_dict.get("ch.ehi.ili2db.beautifyEnumDispName", None)
            == "underscore"
        ):
            self.parameters["beautifyEnumDispName"] = True
        if settings_dict.get("ch.ehi.sqlgen.createGeomIndex", None) == "underscore":
            self.parameters["createGeomIdx"] = True
        if settings_dict.get("ch.ehi.ili2db.createMetaInfo", None):
            self.parameters["createMetaInfo"] = True
        if settings_dict.get("ch.ehi.ili2db.multilingualTrafo", None) == "expand":
            self.parameters["expandMultilingual"] = True
        if settings_dict.get("ch.ehi.ili2db.createTypeConstraint", None):
            self.parameters["createTypeConstraint"] = True
        if settings_dict.get("ch.ehi.ili2db.TidHandling", None) == "property":
            self.parameters["createTidCol"] = True

        return True



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    09/05/19
    git sha              :    :%H$
    copyright            :    (C) 2019 by Yesid Polania (BSF Swissphoto)
    email                :    yesidpol.3@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from qgis.PyQt.QtCore import QSettings

from .db_command_config_manager import DbCommandConfigManager


class MssqlCommandConfigManager(DbCommandConfigManager):

    _settings_base_path = "ili2mssql/"

    def __init__(self, configuration):
        DbCommandConfigManager.__init__(self, configuration)

    def get_uri(self, su=False, qgis=False):
        separator = ";"
        uri = []
        uri += ["DRIVER={{{}}}".format(self.configuration.db_odbc_driver)]
        host = self.configuration.dbhost
        if self.configuration.dbinstance:
            host += "\\" + self.configuration.dbinstance
        if self.configuration.dbport:
            host += "," + self.configuration.dbport

        uri += ["SERVER={}".format(host)]
        uri += ["DATABASE={}".format(self.configuration.database)]
        uri += ["UID={}".format(self.configuration.dbusr)]
        uri += ["PWD={}".format(self.configuration.dbpwd)]

        return separator.join(uri)

    def get_db_args(self, hide_password=False, su=False):
        db_args = list()
        db_args += ["--dbhost", self.configuration.dbhost]
        if self.configuration.dbport:
            db_args += ["--dbport", self.configuration.dbport]
        db_args += ["--dbusr", self.configuration.dbusr]
        if self.configuration.dbpwd:
            if hide_password:
                db_args += ["--dbpwd", "******"]
            else:
                db_args += ["--dbpwd", self.configuration.dbpwd]
        db_args += ["--dbdatabase", self.configuration.database]
        db_args += [
            "--dbschema",
            self.configuration.dbschema or self.configuration.database,
        ]
        if self.configuration.dbinstance:
            db_args += ["--dbinstance", self.configuration.dbinstance]

        return db_args

    def save_config_in_qsettings(self):
        settings = QSettings()

        settings.setValue(self._settings_base_path + "host", self.configuration.dbhost)
        settings.setValue(
            self._settings_base_path + "instance", self.configuration.dbhost
        )
        settings.setValue(self._settings_base_path + "port", self.configuration.dbport)
        settings.setValue(self._settings_base_path + "user", self.configuration.dbusr)
        settings.setValue(
            self._settings_base_path + "database", self.configuration.database
        )
        settings.setValue(
            self._settings_base_path + "schema", self.configuration.dbschema
        )
        settings.setValue(
            self._settings_base_path + "password", self.configuration.dbpwd
        )
        settings.setValue(
            self._settings_base_path + "odbc_driver", self.configuration.db_odbc_driver
        )

    def load_config_from_qsettings(self):
        settings = QSettings()

        self.configuration.dbhost = settings.value(
            self._settings_base_path + "host", "localhost"
        )
        self.configuration.dbport = settings.value(
            self._settings_base_path + "instance"
        )
        self.configuration.dbport = settings.value(self._settings_base_path + "port")
        self.configuration.dbusr = settings.value(self._settings_base_path + "user")
        self.configuration.database = settings.value(
            self._settings_base_path + "database"
        )
        self.configuration.dbschema = settings.value(
            self._settings_base_path + "schema"
        )
        self.configuration.dbpwd = settings.value(self._settings_base_path + "password")
        self.configuration.db_odbc_driver = settings.value(
            self._settings_base_path + "odbc_driver"
        )



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 2022-07-17
        git sha              : :%H$
        copyright            : (C) 2022 by Dave Signer
        email                : david at opengis ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import mimetypes
import os
import uuid
import xml.dom.minidom as minidom
import xml.etree.cElementTree as ET

from .ilitarget import IliTarget


class DatasetMetadata(object):
    """
    Class representing an object of an indexed toppingfile/referencdatafile to be stored in an element DatasetIdx16.DataIndex.DatasetMetadata.
    """

    def __init__(
        self,
        dataset_version: str = None,
        publishing_date: str = None,
        owner: str = None,
        project_name: str = None,
        id: str = None,
        file_type: str = None,
        file_path: str = None,
        linking_models: list = [],
    ):
        self.id = id
        self.file_type = file_type
        self.file_path = file_path
        self.linking_models = linking_models

        self.version = dataset_version
        self.publishing_date = publishing_date
        self.owner = owner
        self.project_name = project_name

        self.title = f"QGIS {self.file_type} file for {self.project_name} - {os.path.splitext(os.path.basename(self.file_path))}"
        self.file_mimetype = self._file_mime_type(self.file_path)

    def _file_mime_type(self, file_path: str = None) -> str:
        mimetype = mimetypes.guess_type(file_path)
        if mimetype[0] is None:
            # ugly fallback
            return "text/plain"
        return mimetype[0]

    def make_xml_element(self, element):
        """
        Generates the content of an element of DatasetIdx16.DataIndex.DatasetMetadata.
        """
        ET.SubElement(element, "id").text = self.id
        ET.SubElement(element, "version").text = self.version
        ET.SubElement(element, "publishingDate").text = self.publishing_date
        ET.SubElement(element, "owner").text = self.owner

        title = ET.SubElement(element, "title")
        datsetidx16_multilingualtext = ET.SubElement(
            title, "DatasetIdx16.MultilingualText"
        )
        localisedtext = ET.SubElement(datsetidx16_multilingualtext, "LocalisedText")
        datasetidx16_localisedtext = ET.SubElement(
            localisedtext, "DatasetIdx16.LocalisedText"
        )
        ET.SubElement(datasetidx16_localisedtext, "Text").text = self.title

        categories = ET.SubElement(element, "categories")
        type_datasetidx16_code = ET.SubElement(categories, "DatasetIdx16.Code_")
        ET.SubElement(
            type_datasetidx16_code, "value"
        ).text = f"http://codes.interlis.ch/type/{self.file_type}"

        if self.file_type in ["metaconfig", "referencedata"]:
            for linking_model in self.linking_models:
                linking_model_datasetidx16_code = ET.SubElement(
                    categories, "DatasetIdx16.Code_"
                )
                ET.SubElement(
                    linking_model_datasetidx16_code, "value"
                ).text = f"http://codes.interlis.ch/model/{linking_model}"

        files = ET.SubElement(element, "files")
        datsaetidx16_datafile = ET.SubElement(files, "DatasetIdx16.DataFile")
        ET.SubElement(datsaetidx16_datafile, "fileFormat").text = self.file_mimetype
        file = ET.SubElement(datsaetidx16_datafile, "file")
        datsetidx16_file = ET.SubElement(file, "DatasetIdx16.File")
        ET.SubElement(datsetidx16_file, "path").text = self.file_path


class IliData(object):
    """
    Class representing the top level content of the ilidata.xml used as "index" for all the toppingfiles/referencedatafile.
    The ilidata.xml is based on the INTERLIS model DatasetIdx16.

    Every indexed toppingfile / referencedatafile is represented by an object of DatasetMetadata.
    """

    def __init__(self):
        pass

    def generate_file(self, target: IliTarget, linking_models: list = []):
        """
        Generates the ilidata.xml file containing all the toppingfile listed in the targets toppingfileinfo_list and writes it to the targets folders.
        """
        transfer = ET.Element("TRANSFER", xmlns="http://www.interlis.ch/INTERLIS2.3")

        headersection = ET.SubElement(
            transfer,
            "HEADERSECTION",
            SENDER="ModelBaker IliProjectTopping",
            VERSION="2.3",
        )
        models = ET.SubElement(headersection, "MODELS")
        ET.SubElement(
            models,
            "MODEL",
            NAME="DatasetIdx16",
            VERSION="2018-11-21",
            URI="mailto:ce@eisenhutinformatik.ch",
        )

        datasection = ET.SubElement(transfer, "DATASECTION")
        data_index = ET.SubElement(
            datasection, "DatasetIdx16.DataIndex", BID=str(uuid.uuid4())
        )

        for toppingfileinfo in target.toppingfileinfo_list:
            dataset_metadata_element = ET.SubElement(
                data_index,
                "DatasetIdx16.DataIndex.DatasetMetadata",
                TID=str(uuid.uuid4()),
            )
            dataset = DatasetMetadata(
                toppingfileinfo.get("version", target.default_version),
                toppingfileinfo.get("publishing_date", target.default_publishing_date),
                toppingfileinfo.get("owner", target.default_owner),
                target.projectname,
                toppingfileinfo["id"],
                toppingfileinfo["type"],
                toppingfileinfo["path"],
                linking_models,
            )
            dataset.make_xml_element(dataset_metadata_element)

        tree = ET.ElementTree(transfer)
        xmlstr = minidom.parseString(ET.tostring(tree.getroot())).toprettyxml(
            indent="   "
        )
        ilidata_path = os.path.join(target.main_dir, "ilidata.xml")
        with open(ilidata_path, "w") as f:
            f.write(xmlstr)
            return ilidata_path



# -*- coding: utf-8 -*-

"""
/***************************************************************************
                              -------------------
        begin                : 2016
        copyright            : (C) 2016 by OPENGIS.ch
        email                : info@opengis.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import fnmatch
import functools
import os.path
import re
import unicodedata
from abc import ABCMeta
from functools import partial

from qgis.core import QgsNetworkAccessManager
from qgis.PyQt.QtCore import (
    QCoreApplication,
    QEventLoop,
    QFile,
    QIODevice,
    QObject,
    QUrl,
)
from qgis.PyQt.QtGui import QValidator
from qgis.PyQt.QtNetwork import QNetworkRequest
from qgis.PyQt.QtWidgets import QApplication, QFileDialog


def selectFileName(line_edit_widget, title, file_filter, parent):
    filename, matched_filter = QFileDialog.getOpenFileName(
        parent, title, line_edit_widget.text(), file_filter
    )
    line_edit_widget.setText(filename)


def make_file_selector(
    widget,
    title=QCoreApplication.translate("modelbaker", "Open File"),
    file_filter=QCoreApplication.translate("modelbaker", "Any file (*)"),
    parent=None,
):
    return partial(
        selectFileName,
        line_edit_widget=widget,
        title=title,
        file_filter=file_filter,
        parent=parent,
    )


def selectFileNameToSave(
    line_edit_widget,
    title,
    file_filter,
    parent,
    extension,
    extensions,
    dont_confirm_overwrite,
):
    filename, matched_filter = QFileDialog.getSaveFileName(
        parent,
        title,
        line_edit_widget.text(),
        file_filter,
        options=QFileDialog.DontConfirmOverwrite
        if dont_confirm_overwrite
        else QFileDialog.Options(),
    )
    extension_valid = False

    if not extensions:
        extensions = [extension]

    if extensions:
        extension_valid = any(filename.endswith(ext) for ext in extensions)

    if not extension_valid and filename:
        filename = filename + extension

    line_edit_widget.setText(filename)


def make_save_file_selector(
    widget,
    title=QCoreApplication.translate("modelbaker", "Open File"),
    file_filter=QCoreApplication.translate("modelbaker", "Any file(*)"),
    parent=None,
    extension="",
    extensions=None,
    dont_confirm_overwrite=False,
):
    return partial(
        selectFileNameToSave,
        line_edit_widget=widget,
        title=title,
        file_filter=file_filter,
        parent=parent,
        extension=extension,
        extensions=extensions,
        dont_confirm_overwrite=dont_confirm_overwrite,
    )


def selectFolder(line_edit_widget, title, parent):
    foldername = QFileDialog.getExistingDirectory(
        parent, title, line_edit_widget.text()
    )
    line_edit_widget.setText(foldername)


def make_folder_selector(
    widget,
    title=QCoreApplication.translate("modelbaker", "Open Folder"),
    parent=None,
):
    return partial(selectFolder, line_edit_widget=widget, title=title, parent=parent)


def slugify(text: str) -> str:
    if not text:
        return text
    slug = unicodedata.normalize("NFKD", text)
    slug = re.sub(r"[^a-zA-Z0-9]+", "_", slug).strip("_")
    slug = re.sub(r"[-]+", "_", slug)
    slug = slug.lower()
    return slug


class NetworkError(RuntimeError):
    def __init__(self, error_code, msg):
        self.msg = msg
        self.error_code = error_code


replies = list()


def download_file(
    url, filename, on_progress=None, on_finished=None, on_error=None, on_success=None
):
    """
    Will download the file from url to a local filename.
    The method will only return once it's finished.

    While downloading it will repeatedly report progress by calling on_progress
    with two parameters bytes_received and bytes_total.

    If an error occurs, it raises a NetworkError exception.

    It will return the filename if everything was ok.
    """
    network_access_manager = QgsNetworkAccessManager.instance()

    req = QNetworkRequest(QUrl(url))
    req.setAttribute(QNetworkRequest.CacheSaveControlAttribute, False)
    req.setAttribute(
        QNetworkRequest.CacheLoadControlAttribute, QNetworkRequest.AlwaysNetwork
    )
    req.setAttribute(QNetworkRequest.FollowRedirectsAttribute, True)
    reply = network_access_manager.get(req)

    def on_download_progress(bytes_received, bytes_total):
        on_progress(bytes_received, bytes_total)

    def finished(filename, reply, on_error, on_success, on_finished):
        file = QFile(filename)
        file.open(QIODevice.WriteOnly)
        file.write(reply.readAll())
        file.close()
        if reply.error() and on_error:
            on_error(reply.error(), reply.errorString())
        elif not reply.error() and on_success:
            on_success()

        if on_finished:
            on_finished()
        reply.deleteLater()
        replies.remove(reply)

    if on_progress:
        reply.downloadProgress.connect(on_download_progress)

    on_reply_finished = functools.partial(
        finished, filename, reply, on_error, on_success, on_finished
    )

    reply.finished.connect(on_reply_finished)

    replies.append(reply)

    if not on_finished and not on_success:
        loop = QEventLoop()
        reply.finished.connect(loop.quit)
        loop.exec_()

        if reply.error():
            raise NetworkError(reply.error(), reply.errorString())
        else:
            return filename


class Validators(QObject):
    def validate_line_edits(self, *args, **kwargs):
        """
        Validate line edits and set their color to indicate validation state.
        """
        senderObj = self.sender()
        validator = senderObj.validator()
        if validator is None:
            color = "#fff"  # White
        else:
            state = validator.validate(senderObj.text().strip(), 0)[0]
            if state == QValidator.Acceptable:
                color = "#fff"  # White
            elif state == QValidator.Intermediate:
                color = "#ffd356"  # Light orange
            else:
                color = "#f6989d"  # Red
        senderObj.setStyleSheet("QLineEdit {{ background-color: {} }}".format(color))


class FileValidator(QValidator):
    def __init__(
        self,
        pattern="*",
        is_executable=False,
        parent=None,
        allow_empty=False,
        allow_non_existing=False,
    ):
        """
        Validates if a string is a valid filename, based on the provided parameters.

        :param pattern: A file glob pattern as recognized by ``fnmatch``, if a list if provided, the validator will try
                        to match every pattern in the list.
        :param is_executable: Only match executable files
        :param parent: The parent QObject
        :param allow_empty: Empty strings are valid
        :param allow_non_existing: Non existing files are valid
        """
        QValidator.__init__(self, parent)
        self.pattern = pattern
        self.is_executable = is_executable
        self.allow_empty = allow_empty
        self.allow_non_existing = allow_non_existing
        self.error = ""

    """
    Validator for file line edits
    """

    def validate(self, text, pos):
        self.error = ""

        if self.allow_empty and not text.strip():
            return QValidator.Acceptable, text, pos

        pattern_matches = False
        if type(self.pattern) is str:
            pattern_matches = fnmatch.fnmatch(text, self.pattern)
        elif type(self.pattern) is list:
            pattern_matches = True in (
                fnmatch.fnmatch(text, pattern) for pattern in self.pattern
            )
        else:
            raise TypeError(
                "pattern must be str or list, not {}".format(type(self.pattern))
            )

        if not text:
            self.error = self.tr("Text field value is empty.")
        elif not self.allow_non_existing and not os.path.isfile(text):
            self.error = self.tr("The chosen file does not exist.")
        elif not pattern_matches:
            self.error = self.tr(
                "The chosen file has a wrong extension (has to be {}).".format(
                    self.pattern
                    if type(self.pattern) is str
                    else ",".join(self.pattern)
                )
            )
        elif self.is_executable and not os.access(text, os.X_OK):
            self.error = self.tr("The chosen file is not executable.")
        if self.error:
            return QValidator.Intermediate, text, pos
        else:
            return QValidator.Acceptable, text, pos


class NonEmptyStringValidator(QValidator):
    def __init__(self, parent=None):
        QValidator.__init__(self, parent)

    def validate(self, text, pos):
        if not text.strip():
            return QValidator.Intermediate, text, pos

        return QValidator.Acceptable, text, pos


class OverrideCursor:
    def __init__(self, cursor):
        self.cursor = cursor

    def __enter__(self):
        QApplication.setOverrideCursor(self.cursor)

    def __exit__(self, exc_type, exc_val, exc_tb):
        QApplication.restoreOverrideCursor()


class AbstractQObjectMeta(ABCMeta, type(QObject)):
    """Metaclass that is used by any class that must be abstract and inherit from QObject.

    If a class must be abstract (ABC or ABCMeta) and inherit from QObject, multiple inheritance fails because
    the classes have different metaclasses. See more:

    https://stackoverflow.com/questions/46837947/how-to-create-an-abstract-base-class-in-python-which-derived-from-qobject
    """



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    10/05/19
    git sha              :    :%H$
    copyright            :    (C) 2019 by Yesid Polania
    email                :    yesidpol.3@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from ..dataobjects.fields import Field
from ..dbconnector.mssql_connector import MssqlConnector
from .db_factory import DbFactory
from .mssql_command_config_manager import MssqlCommandConfigManager
from .mssql_layer_uri import MssqlLayerUri


class MssqlFactory(DbFactory):
    def get_db_connector(self, uri, schema):
        return MssqlConnector(uri, schema)

    def get_db_command_config_manager(self, configuration):
        return MssqlCommandConfigManager(configuration)

    def get_layer_uri(self, uri):
        return MssqlLayerUri(uri)

    def pre_generate_project(self, configuration):
        return True, ""

    def post_generate_project_validations(self, configuration):
        return True, ""

    def customize_widget_editor(self, field: Field, data_type: str):
        if "bit" in data_type:
            field.widget = "CheckBox"
            field.widget_config["CheckedState"] = "1"
            field.widget_config["UncheckedState"] = "0"



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 11/11/21
        git sha              : :%H$
        copyright            : (C) 2021 by Dave Signer
        email                : david at opengis ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import xml.etree.cElementTree as CET
from enum import Enum

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QStandardItem, QStandardItemModel

from .ili2dbconfig import ValidateConfiguration
from .iliexecutable import IliExecutable


class Validator(IliExecutable):
    def __init__(self, parent=None):
        super(Validator, self).__init__(parent)

    def _create_config(self):
        return ValidateConfiguration()


class ValidationResultModel(QStandardItemModel):
    """
    Model containing all the error/warning data of the current xtf file.
    """

    class Roles(Enum):
        ID = Qt.UserRole + 1
        MESSAGE = Qt.UserRole + 2
        TYPE = Qt.UserRole + 3
        OBJ_TAG = Qt.UserRole + 4
        TID = Qt.UserRole + 5
        TECH_ID = Qt.UserRole + 6
        USER_ID = Qt.UserRole + 7
        ILI_Q_NAME = Qt.UserRole + 8
        DATA_SOURCE = Qt.UserRole + 9
        LINE = Qt.UserRole + 10
        COORD_X = Qt.UserRole + 11
        COORD_Y = Qt.UserRole + 12
        TECH_DETAILS = Qt.UserRole + 13

        FIXED = Qt.UserRole + 14

        def __int__(self):
            return self.value

    def __init__(self):
        super().__init__()
        self.configuration = ValidateConfiguration()
        self.valid = False

    def get_element_text(self, element):
        if element is not None:
            return element.text
        return None

    def reload(self):
        self.beginResetModel()
        if self.configuration.xtflog:
            try:
                root = CET.parse(self.configuration.xtflog).getroot()
            except CET.ParseError as e:
                print(
                    self.tr(
                        "Could not parse ilidata file `{file}` ({exception})".format(
                            file=self.configuration.xtflog, exception=str(e)
                        )
                    )
                )
            if root:
                ns = "{http://www.interlis.ch/INTERLIS2.3}"
                for error in root.iter(ns + "IliVErrors.ErrorLog.Error"):
                    id = error.attrib["TID"]
                    message = self.get_element_text(error.find(ns + "Message"))
                    type = self.get_element_text(error.find(ns + "Type"))
                    obj_tag = self.get_element_text(error.find(ns + "ObjTag"))
                    tid = self.get_element_text(error.find(ns + "Tid"))
                    tech_id = self.get_element_text(error.find(ns + "TechId"))
                    user_id = self.get_element_text(error.find(ns + "UserId"))
                    ili_q_name = self.get_element_text(error.find(ns + "IliQName"))
                    data_source = self.get_element_text(error.find(ns + "DataSource"))
                    line = self.get_element_text(error.find(ns + "Line"))
                    coord_x = None
                    coord_y = None
                    geometry = error.find(ns + "Geometry")
                    if geometry:
                        coord = geometry.find(ns + "COORD")
                        if coord:
                            coord_x = self.get_element_text(coord.find(ns + "C1"))
                            coord_y = self.get_element_text(coord.find(ns + "C2"))
                    tech_details = self.get_element_text(error.find(ns + "TechDetails"))

                    if type in ["Error", "Warning"] and message != "...validate failed":
                        item = QStandardItem()
                        item.setData(id, int(ValidationResultModel.Roles.ID))
                        item.setData(message, int(ValidationResultModel.Roles.MESSAGE))
                        item.setData(type, int(ValidationResultModel.Roles.TYPE))
                        item.setData(obj_tag, int(ValidationResultModel.Roles.OBJ_TAG))
                        item.setData(tid, int(ValidationResultModel.Roles.TID))
                        item.setData(tech_id, int(ValidationResultModel.Roles.TECH_ID))
                        item.setData(user_id, int(ValidationResultModel.Roles.USER_ID))
                        item.setData(
                            ili_q_name, int(ValidationResultModel.Roles.ILI_Q_NAME)
                        )
                        item.setData(
                            data_source, int(ValidationResultModel.Roles.DATA_SOURCE)
                        )
                        item.setData(line, int(ValidationResultModel.Roles.LINE))
                        item.setData(coord_x, int(ValidationResultModel.Roles.COORD_X))
                        item.setData(coord_y, int(ValidationResultModel.Roles.COORD_Y))
                        item.setData(
                            tech_details, int(ValidationResultModel.Roles.TECH_DETAILS)
                        )
                        item.setData(False, int(ValidationResultModel.Roles.FIXED))
                        self.appendRow(item)
        self.endResetModel()



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 2017-04-12
        git sha              : :%H$
        copyright            : (C) 2017 by OPENGIS.ch
        email                : info@opengis.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.core import QgsDefaultValue, QgsEditorWidgetSetup


class Field:
    def __init__(self, name):
        self.name = name
        self.alias = None
        self.read_only = False
        self.widget = None
        self.widget_config = dict()
        self.default_value_expression = None
        self.enum_domain = None
        self.oid_domain = None

    def dump(self):
        definition = dict()
        if self.alias:
            definition["alias"] = self.alias

        return definition

    def load(self, definition):
        if "alias" in definition:
            self.alias = definition["alias"]

    def create(self, layer):
        field_idx = layer.layer.fields().indexOf(self.name)

        if self.alias:
            layer.layer.setFieldAlias(field_idx, self.alias)

        if self.widget:
            setup = QgsEditorWidgetSetup(self.widget, self.widget_config)
            layer.layer.setEditorWidgetSetup(field_idx, setup)

        if self.default_value_expression:
            default_value = QgsDefaultValue(self.default_value_expression)
            layer.layer.setDefaultValueDefinition(field_idx, default_value)



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    04/10/17
    git sha              :    :%H$
    copyright            :    (C) 2017 by Germán Carrillo (BSF-Swissphoto)
    email                :    gcarrillo@linuxmail.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.PyQt.QtCore import QObject, pyqtSignal

from .config import BASKET_TABLES, IGNORED_ILI_ELEMENTS, IGNORED_SCHEMAS, IGNORED_TABLES


class DBConnector(QObject):
    """SuperClass for all DB connectors."""

    stdout = pyqtSignal(str)
    new_message = pyqtSignal(int, str)

    def __init__(self, uri, schema, parent=None):
        QObject.__init__(self, parent)
        self.QGIS_DATE_TYPE = "date"
        self.QGIS_TIME_TYPE = "time"
        self.QGIS_DATE_TIME_TYPE = "datetime"
        self.iliCodeName = ""  # For Domain-Class relations, specific for each DB
        self.tid = ""  # For BAG OF config and basket handling, specific for each DB
        self.tilitid = ""  # For basket handling, specific for each DB
        self.dispName = ""  # For BAG OF config, specific for each DB
        self.basket_table_name = ""  # For basket handling, specific for each DB
        self.dataset_table_name = ""  # For basket handling, specific for each DB

    def map_data_types(self, data_type):
        """Map provider date/time types to QGIS date/time types"""
        return None

    def db_or_schema_exists(self):
        """Whether the DB (for GPKG) or schema (for PG) exists or not."""
        raise NotImplementedError

    def create_db_or_schema(self, usr):
        """Create the DB (for GPKG) or schema (for PG)"""
        raise NotImplementedError

    def metadata_exists(self):
        """Whether t_ili2db_table_prop table exists or not.
        In other words... Does the DB/Schema hold an Interlis model?
        """
        return False

    def get_tables_info(self):
        """
        Info about tables found in the database (or database schema).

        Return:
            Iterable allowing to access rows, each row should allow to access
            specific columns by name (e.g., a list of dicts {column_name:value})
            Expected columns are:
                schemaname
                tablename
                primary_key
                geometry_column
                srid
                type  (Geometry Type)
                kind_settings
                ili_name
                extent [a string: "xmin;ymin;xmax;ymax"]
                table_alias
                model
        """
        return []

    def get_meta_attrs_info(self):
        """
        Info about meta attributes

        Return:
            Iterable allowing to access rows, each row should allow to access
            specific columns by name (e.g., a list of dicts {column_name:value})
            Expected columns are:
                ilielement
                attr_name
                attr_value
        """
        raise NotImplementedError

    def get_meta_attr(self, ili_name):
        """
        Info about meta attributes of a given ili element

        Return:
            Iterable allowing to access rows, each row should allow to access
            specific columns by name (e.g., a list of dicts {column_name:value})
            Expected columns are:
                attr_name
                attr_value
        """
        return []

    def get_fields_info(self, table_name):
        """
        Info about fields of a given table in the database.

        Return:
            Iterable allowing to access rows, each row should allow to access
            specific columns by name (e.g., a list of dicts {column_name:value})
            Expected columns are:
                column_name
                data_type
                comment
                unit
                texttype
                column_alias
                enum_domain
                oid_domain
                fully_qualified_name
        """
        return []

    def get_min_max_info(self, table_name):
        """
        Info about range constraints found in a given table.

        Return:
            Dictionary with keys corresponding to column names and values
            corresponding to tuples in the form (min_value, max_value)
        """
        return {}

    def get_value_map_info(self, table_name):
        """
        Info about value map constraints found in a given table.

        Return:
            Dictionary with keys corresponding to column names and values
            with lists of allowed values
        """
        return {}

    def get_relations_info(self, filter_layer_list=[]):
        """
        Info about relations found in a database (or database schema).

        Return:
            Iterable allowing to access rows, each row should allow to access
            specific columns by name (e.g., a list of dicts {column_name:value})
            Expected columns are:
                constraint_name
                referencing_table_name
                referencing_column_name
                constraint_schema
                referenced_table_name
                referenced_column_name
                strength
                cardinality_max
        """
        return []

    def get_bags_of_info(self):
        """
        Info about bags_of found in a database (or database schema).

        Return:
            Iterable allowing to access rows, each row should allow to access
            specific columns by name (e.g., a list of dicts {column_name:value})
            Expected columns are:
                current_layer_name
                attribute
                target_layer_name
                cardinality_max
                cardinality_min
        """
        return []

    def get_ignored_layers(self, ignore_basket_tables=True):
        """
        The ignored layers according to the ignored schemas and ignored tables and the ignored ili elements
        listed in the config.py.
        Additionally all the ili elements that have the attribute name ili2db.mapping in the meta attribute
        table.
        """
        tables_info = self.get_tables_info()
        relations_info = self.get_relations_info()
        meta_attrs_info = self.get_meta_attrs_info()
        mapping_ili_elements = []
        static_tables = []
        detected_tables = []
        referencing_detected_tables = []
        for record in meta_attrs_info:
            if record["attr_name"] == "ili2db.mapping":
                mapping_ili_elements.append(record["ilielement"])

        for record in tables_info:
            if "ili_name" in record:
                if (
                    record["ili_name"] in mapping_ili_elements
                    or record["ili_name"] in IGNORED_ILI_ELEMENTS
                ):
                    detected_tables.append(record["tablename"])
                    continue
            if "schemaname" in record:
                if record["schemaname"] in IGNORED_SCHEMAS:
                    static_tables.append(record["tablename"])
                    continue
            if "tablename" in record:
                if self.is_spatial_index_table(record["tablename"]) or (
                    record["tablename"] in IGNORED_TABLES
                    and (
                        ignore_basket_tables or record["tablename"] not in BASKET_TABLES
                    )
                ):
                    static_tables.append(record["tablename"])
                    continue

        for record in relations_info:
            if record["referenced_table"] in detected_tables:
                referencing_detected_tables.append(record["referencing_table"])

        return static_tables + detected_tables + referencing_detected_tables

    def is_spatial_index_table(self, tablename=str) -> bool:
        """Note: Checks if the table is a technical table used for spatial indexing"""
        return False

    def get_iliname_dbname_mapping(self, sqlnames=list()):
        """Note: the parameter sqlnames is only used for ili2db version 3 relation creation"""
        return {}

    def get_classili_classdb_mapping(self, models_info, extended_classes):
        """Used for ili2db version 3"""
        return {}

    def get_attrili_attrdb_mapping(self, attrs_list):
        """Used for ili2db version 3"""
        return {}

    def get_attrili_attrdb_mapping_by_owner(self, owners):
        """Used for ili2db version 3"""
        return {}

    def get_models(self):
        return {}

    def ili_version(self):
        """
        Returns the version of the ili2db application that was used to create the schema
        """
        return None

    def get_basket_handling(self):
        """
        Returns `True` if a basket handling is enabled according to the settings table.
        Means when the database has been created with `--createBasketCol`.
        """
        return False

    def get_baskets_info(self):
        """
        Info about baskets found in the basket table and the related datasetname
        Return:
            Iterable allowing to access rows, each row should allow to access
            specific columns by name (e.g., a list of dicts {column_name:value})
            Expected columns are:
                basket_t_id
                basket_t_ili_tid
                topic
                dataset_t_id
                datasetname
        """
        return {}

    def get_datasets_info(self):
        """
        Info about datasets found in the dataset table
        Return:
            Iterable allowing to access rows, each row should allow to access
            specific columns by name (e.g., a list of dicts {column_name:value})
            Expected columns are:
                t_id
                datasetname
        """
        return {}

    def create_dataset(self, datasetname):
        """
        Returns the state and the errormessage
        """
        return False, None

    def rename_dataset(self, tid, datasetname):
        """
        Returns the state and the errormessage
        """
        return False, None

    def get_topics_info(self):
        """
        Returns all the topics found in the table t_ili2db_classname as long as they contain tables found in t_ili2db_table_prop and are there not defined as ENUM.
        This avoids to get structures back, containing enumerations and not being in a topic, having the structure <modelname>.<structurename>.<enumerationname>.
        Return:
            Iterable allowing to access rows, each row should allow to access
            specific columns by name (e.g., a list of dicts {column_name:value})
            Expected columns are:
                model
                topic
        """
        return {}

    def create_basket(self, dataset_tid, topic):
        """
        Returns the state and the errormessage
        """
        return False, None

    def get_tid_handling(self):
        """
        Returns `True` if a tid handling is enabled according to the settings table (when the database has been created with `--createTidCol`).
        If t_ili_tids are used only because of a stable id definition in the model (with `OID as` in the topic or the class definition), this parameter is not set and this function will return `{}`.
        """
        return {}

    def get_ili2db_settings(self):
        """
        Returns the settings like they are without any name mapping etc.
        """
        return {}


class DBConnectorError(Exception):
    """This error is raised when DbConnector could not connect to database.

    This exception wraps different database exceptions to unify them in a single exception.
    """

    def __init__(self, message, base_exception=None):
        super().__init__(message)
        self.base_exception = base_exception



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    01/02/19
    git sha              :    :%H$
    copyright            :    (C) 2019 by Yesid Polanía (BSF-Swissphoto)
    email                :    yesidpol.3@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import re

import pyodbc
from qgis.core import Qgis

from .db_connector import DBConnector, DBConnectorError

METADATA_TABLE = "t_ili2db_table_prop"
METAATTRS_TABLE = "t_ili2db_meta_attrs"
SETTINGS_TABLE = "t_ili2db_settings"
DATASET_TABLE = "t_ili2db_dataset"
BASKET_TABLE = "t_ili2db_basket"


class MssqlConnector(DBConnector):
    def __init__(self, uri, schema):
        DBConnector.__init__(self, uri, schema)

        try:
            self.conn = pyodbc.connect(uri)
        except (
            pyodbc.ProgrammingError,
            pyodbc.InterfaceError,
            pyodbc.Error,
            pyodbc.OperationalError,
        ) as e:
            raise DBConnectorError(str(e), e)

        self.schema = schema

        self._bMetadataTable = self._metadata_exists()
        self.iliCodeName = "iliCode"
        self.tid = "T_Id"
        self.tilitid = "T_Ili_Tid"
        self.dispName = "dispName"
        self.basket_table_name = BASKET_TABLE
        self.dataset_table_name = DATASET_TABLE

    def map_data_types(self, data_type):
        result = data_type.lower()
        if "timestamp" in data_type:
            result = self.QGIS_DATE_TIME_TYPE
        elif "date" in data_type:
            result = self.QGIS_DATE_TYPE
        elif "time" in data_type:
            result = self.QGIS_TIME_TYPE

        return result

    def db_or_schema_exists(self):
        if self.schema:
            cur = self.conn.cursor()
            cur.execute(
                """
                SELECT case when count(schema_name)>0 then 1 else 0 end
                FROM information_schema.schemata
                where schema_name = '{}'
            """.format(
                    self.schema
                )
            )

            return bool(cur.fetchone()[0])

        return False

    def metadata_exists(self):
        return self._bMetadataTable

    def _metadata_exists(self):
        return self._table_exists(METADATA_TABLE)

    def _table_exists(self, tablename):
        if self.schema:
            cur = self.conn.cursor()
            cur.execute(
                """
            SELECT count(TABLE_NAME) as 'count'
                FROM INFORMATION_SCHEMA.TABLES
                WHERE TABLE_TYPE = 'BASE TABLE'
                AND TABLE_SCHEMA = '{}'
                    AND TABLE_NAME = '{}'
            """.format(
                    self.schema, tablename
                )
            )

            return bool(cur.fetchone()[0])

        return False

    def get_tables_info(self):
        res = []

        if self.schema:
            metadata_exists = self.metadata_exists()

            ln = "\n"
            stmt = ""
            stmt += ln + "SELECT distinct"
            stmt += ln + "     tbls.TABLE_SCHEMA AS schemaname"
            stmt += ln + "    , tbls.TABLE_NAME AS tablename"
            stmt += ln + "    , Col.Column_Name AS primary_key"
            stmt += ln + "    , clm.COLUMN_NAME AS geometry_column"
            if metadata_exists:
                stmt += ln + "    , tsrid.setting AS srid"
                stmt += ln + "    , p.setting AS kind_settings"
                stmt += ln + "    , alias.setting AS table_alias"
                stmt += (
                    ln + "    , left(c.iliname, charindex('.', c.iliname)-1) AS model"
                )
                stmt += ln + "    , c.iliname AS ili_name"
                stmt += ln + "    , STUFF("
                stmt += ln + "       (SELECT ';' + CAST(cp.setting AS VARCHAR(MAX))"
                stmt += ln + "        FROM {schema}.t_ili2db_column_prop cp"
                stmt += ln + "        WHERE tbls.table_name = cp.tablename"
                stmt += ln + "            AND clm.COLUMN_NAME = cp.columnname"
                stmt += ln + "            AND cp.tag IN"
                stmt += (
                    ln
                    + "                ('ch.ehi.ili2db.c1Min', 'ch.ehi.ili2db.c2Min',"
                )
                stmt += (
                    ln + "                'ch.ehi.ili2db.c1Max', 'ch.ehi.ili2db.c2Max')"
                )
                stmt += (
                    ln
                    + "        order by case cp.tag WHEN 'ch.ehi.ili2db.c1Min' THEN 1"
                )
                stmt += ln + "            WHEN 'ch.ehi.ili2db.c2Min' THEN 2"
                stmt += ln + "            WHEN 'ch.ehi.ili2db.c1Max' THEN 3"
                stmt += ln + "            WHEN 'ch.ehi.ili2db.c2Max' THEN 4"
                stmt += ln + "            END"
                stmt += (
                    ln
                    + "        FOR XML PATH(''),TYPE).value('(./text())[1]','VARCHAR(MAX)'),1,1,''"
                )
                stmt += ln + "        ) AS extent"
                stmt += (
                    ln
                    + "    , ( SELECT CASE MAX(CHARINDEX('.',cp.setting)) WHEN 0 THEN 0 ELSE MAX( LEN(cp.setting) -  CHARINDEX('.',cp.setting) ) END"
                )
                stmt += ln + "        FROM {schema}.t_ili2db_column_prop cp"
                stmt += ln + "        WHERE tbls.table_name = cp.tablename"
                stmt += ln + "            AND clm.COLUMN_NAME = cp.columnname"
                stmt += ln + "            AND cp.tag IN"
                stmt += (
                    ln
                    + "                ('ch.ehi.ili2db.c1Min', 'ch.ehi.ili2db.c2Min',"
                )
                stmt += (
                    ln + "                'ch.ehi.ili2db.c1Max', 'ch.ehi.ili2db.c2Max')"
                )
                stmt += ln + "      ) AS coord_decimals"
                stmt += ln + "    , tgeomtype.setting AS simple_type"
                stmt += ln + "    , null AS formatted_type"
                stmt += ln + "    , attrs.sqlname AS attribute_name"
            stmt += ln + "FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS AS Tab"
            stmt += ln + "INNER JOIN INFORMATION_SCHEMA.CONSTRAINT_COLUMN_USAGE AS Col"
            stmt += ln + "    ON Col.Constraint_Name = Tab.Constraint_Name"
            stmt += ln + "    AND Col.Table_Name = Tab.Table_Name"
            stmt += ln + "    AND Col.CONSTRAINT_SCHEMA = Tab.CONSTRAINT_SCHEMA"
            stmt += ln + "RIGHT JOIN INFORMATION_SCHEMA.TABLES AS tbls"
            stmt += ln + "    ON Tab.TABLE_NAME = tbls.TABLE_NAME"
            stmt += ln + "    AND Tab.CONSTRAINT_SCHEMA = tbls.TABLE_SCHEMA"
            stmt += ln + "    AND Tab.Constraint_Type = 'PRIMARY KEY'"
            if metadata_exists:
                stmt += ln + "LEFT JOIN {schema}.T_ILI2DB_TABLE_PROP AS p"
                stmt += ln + "    ON p.tablename = tbls.TABLE_NAME"
                stmt += ln + "    AND p.tag = 'ch.ehi.ili2db.tableKind'"
                stmt += ln + "LEFT JOIN {schema}.T_ILI2DB_TABLE_PROP AS alias"
                stmt += ln + "    ON alias.tablename = tbls.TABLE_NAME"
                stmt += ln + "    AND alias.tag = 'ch.ehi.ili2db.dispName'"
                stmt += ln + "LEFT JOIN {schema}.t_ili2db_classname AS c"
                stmt += ln + "    ON tbls.TABLE_NAME = c.sqlname"
                stmt += ln + "LEFT JOIN {schema}.t_ili2db_attrname AS attrs"
                stmt += ln + "    ON c.iliname = attrs.iliname"
            stmt += ln + "LEFT JOIN INFORMATION_SCHEMA.COLUMNS AS clm"
            stmt += ln + "    ON clm.TABLE_NAME = tbls.TABLE_NAME"
            stmt += ln + "    AND clm.TABLE_SCHEMA = tbls.TABLE_SCHEMA"
            stmt += ln + "    AND clm.DATA_TYPE = 'geometry'"
            if metadata_exists:
                stmt += ln + "LEFT JOIN {schema}.T_ILI2DB_COLUMN_PROP AS tsrid"
                stmt += ln + "    ON tbls.TABLE_NAME = tsrid.tablename"
                stmt += ln + "    AND clm.COLUMN_NAME = tsrid.columnname"
                stmt += ln + "    AND tsrid.tag='ch.ehi.ili2db.srid'"
                stmt += ln + "LEFT JOIN {schema}.T_ILI2DB_COLUMN_PROP AS tgeomtype"
                stmt += ln + "    ON tbls.TABLE_NAME = tgeomtype.tablename"
                stmt += ln + "    AND clm.COLUMN_NAME = tgeomtype.columnname"
                stmt += ln + "    AND tgeomtype.tag= 'ch.ehi.ili2db.geomType'"
            stmt += (
                ln
                + "WHERE tbls.TABLE_TYPE = 'BASE TABLE' AND tbls.TABLE_SCHEMA = '{schema}'"
            )
            stmt = stmt.format(schema=self.schema)

            if not metadata_exists:
                stmt = self._def_cursor(stmt)

            cur = self.conn.cursor()
            cur.execute(stmt)

            if cur.description:
                columns = [column[0] for column in cur.description]

                for row in cur.fetchall():
                    my_rec = dict(zip(columns, row))
                    my_rec["srid"] = int(my_rec["srid"]) if my_rec["srid"] else None
                    my_rec["type"] = my_rec["simple_type"]
                    res.append(my_rec)

        return res

    def _def_cursor(self, query):
        cursor = """
            DECLARE @schemaname VARCHAR(1000)
            DECLARE @tablename VARCHAR(1000)
            DECLARE @primary_key VARCHAR(1000)
            DECLARE @geometry_column VARCHAR(1000)
            DECLARE @query NVARCHAR(MAX)
            DECLARE @row_counter NVARCHAR(MAX)
            DECLARE @separator NVARCHAR(200)
            declare @full_tabname nvarchar(300)
            declare @count int
            set @query = ''
            set @row_counter = ''
            set @separator = ''

            DECLARE db_cursor CURSOR FOR {query}
            OPEN db_cursor
            FETCH NEXT FROM db_cursor INTO @schemaname, @tablename, @primary_key, @geometry_column
            WHILE @@FETCH_STATUS = 0
            BEGIN
                if @schemaname is null	set @full_tabname = @tablename else set @full_tabname = concat(@schemaname,'.', @tablename)
                if @schemaname is not null set @schemaname = CONCAT('''', @schemaname, '''') else set @schemaname = 'NULL'
                if @primary_key is not null set @primary_key = CONCAT('''', @primary_key, '''') else set @primary_key = 'NULL'
                set @tablename = CONCAT('''', @tablename, '''')

                if @geometry_column is not null begin -- table has geometry
                    select @row_counter = concat('SELECT @C=count(concat(',@geometry_column,'.STSrid,', @geometry_column,'.STGeometryType()))', ' from ', @full_tabname)
                    execute sp_executesql @row_counter, N'@C INT OUTPUT', @C=@count OUTPUT

                    if @count > 0 begin -- table containts geometry data, geometry types are got from data
                        set @query = concat('SELECT distinct '
                        , @schemaname, ' as schemaname,'
                        , @tablename, ' as tablename,'
                        , @primary_key, ' as primary_key,'
                        , '''', @geometry_column, ''' as geometry_column,'
                        , @geometry_column,'.STSrid as srid,'
                        , @geometry_column,'.STGeometryType() as simple_type'
                        , ' from ', @full_tabname, @separator, @query)
                    end else begin -- otherwise, geometry type is set null
                        set @query = concat('SELECT '
                        , @schemaname, ' as schemaname,'
                        , @tablename, ' as tablename,'
                        , @primary_key, ' as primary_key,'
                        , '''', @geometry_column, ''' as geometry_column,'
                        , 'NULL as srid, NULL as simple_type'
                        , @separator, @query)
                    end
                end	else begin -- table does not have geometry
                    set @query = concat('SELECT '
                        , @schemaname, ' as schemaname,'
                        , @tablename, ' as tablename,'
                        , @primary_key, ' as primary_key,'
                        , N'NULL as geometry_column'
                        , N', NULL as srid, NULL as simple_type'
                        , @separator, @query)
                end
                set @separator = ' union '
                FETCH NEXT FROM db_cursor INTO @schemaname, @tablename, @primary_key, @geometry_column
            END
            CLOSE db_cursor
            DEALLOCATE db_cursor
            execute sp_executesql @query """.format(
            query=query
        )

        return cursor

    def get_meta_attrs_info(self):
        if not self._table_exists(METAATTRS_TABLE):
            return []

        result = []

        if self.schema:
            cur = self.conn.cursor()
            cur.execute(
                """
                        SELECT *
                        FROM {schema}.{metaattrs_table};
            """.format(
                    schema=self.schema,
                    metaattrs_table=METAATTRS_TABLE,
                )
            )

            result = self._get_dict_result(cur)

        return result

    def get_meta_attrs(self, ili_name):
        if not self._table_exists(METAATTRS_TABLE):
            return []

        result = []

        if self.schema:
            cur = self.conn.cursor()
            cur.execute(
                """
                        SELECT
                          attr_name,
                          attr_value
                        FROM {schema}.{metaattrs_table}
                        WHERE ilielement='{ili_name}';
            """.format(
                    schema=self.schema,
                    metaattrs_table=METAATTRS_TABLE,
                    ili_name=ili_name,
                )
            )

            result = self._get_dict_result(cur)

        return result

    def get_fields_info(self, table_name):
        res = []
        # Get all fields for this table
        if self.schema:
            metadata_exists = self.metadata_exists()
            metaattrs_exists = self._table_exists(METAATTRS_TABLE)
            ln = "\n"
            stmt = ""

            # TODO description column is missing
            stmt += ln + "SELECT"
            stmt += ln + "     c.column_name"
            stmt += (
                ln
                + "    , case c.data_type when 'decimal' then 'numeric' else c.DATA_TYPE end as data_type"
            )
            stmt += ln + "    , c.numeric_scale"
            if metadata_exists:
                stmt += ln + "    , unit.setting AS unit"
                stmt += ln + "    , txttype.setting AS texttype"
                stmt += ln + "    , alias.setting AS column_alias"
                stmt += ln + "    , full_name.iliname AS fully_qualified_name"
                stmt += ln + "    , enum_domain.setting AS enum_domain"
                stmt += ln + "    , oid_domain.setting AS oid_domain"
                if metaattrs_exists:
                    stmt += (
                        ln
                        + "    , COALESCE(CAST(form_order.attr_value AS int), 999) AS attr_order"
                        + "    , attr_mapping.attr_value AS attr_mapping"
                    )
            stmt += ln + "    , null AS comment"
            stmt += ln + "FROM INFORMATION_SCHEMA.COLUMNS AS c"
            if metadata_exists:
                stmt += ln + "LEFT JOIN {schema}.t_ili2db_column_prop unit"
                stmt += ln + "    ON c.table_name = unit.tablename"
                stmt += ln + "    AND c.column_name = unit.columnname"
                stmt += ln + "    AND unit.tag = 'ch.ehi.ili2db.unit'"
                stmt += ln + "LEFT JOIN {schema}.t_ili2db_column_prop txttype"
                stmt += ln + "    ON c.table_name = txttype.tablename"
                stmt += ln + "    AND c.column_name = txttype.columnname"
                stmt += ln + "    AND txttype.tag = 'ch.ehi.ili2db.textKind'"
                stmt += ln + "LEFT JOIN {schema}.t_ili2db_column_prop alias"
                stmt += ln + "    ON c.table_name = alias.tablename"
                stmt += ln + "    AND c.column_name = alias.columnname"
                stmt += ln + "    AND alias.tag = 'ch.ehi.ili2db.dispName'"
                stmt += ln + "LEFT JOIN {schema}.t_ili2db_attrname full_name"
                stmt += ln + "    ON full_name.{}='{{table}}'".format(
                    "owner" if self.ili_version() == 3 else "colowner"
                )
                stmt += ln + "    AND c.column_name=full_name.sqlname"
                stmt += ln + "LEFT JOIN {schema}.t_ili2db_column_prop enum_domain"
                stmt += ln + "    ON c.table_name = enum_domain.tablename"
                stmt += ln + "    AND c.column_name = enum_domain.columnname"
                stmt += ln + "    AND enum_domain.tag = 'ch.ehi.ili2db.enumDomain'"
                stmt += ln + "LEFT JOIN {schema}.t_ili2db_column_prop oid_domain"
                stmt += ln + "    ON c.table_name = oid_domain.tablename"
                stmt += (
                    ln + "    AND LOWER(c.column_name) = LOWER(oid_domain.columnname)"
                )
                stmt += ln + "    AND oid_domain.tag = 'ch.ehi.ili2db.oidDomain'"
                if metaattrs_exists:
                    stmt += ln + "LEFT JOIN {schema}.t_ili2db_meta_attrs form_order"
                    stmt += ln + "    ON full_name.iliname=form_order.ilielement AND"
                    stmt += ln + "    form_order.attr_name='form_order'"
                    stmt += ln + "LEFT JOIN {schema}.t_ili2db_meta_attrs attr_mapping"
                    stmt += ln + "    ON full_name.iliname=attr_mapping.ilielement AND"
                    stmt += ln + "    attr_mapping.attr_name='ili2db.mapping'"
            stmt += ln + "WHERE TABLE_NAME = '{table}' AND TABLE_SCHEMA = '{schema}'"
            if metadata_exists and metaattrs_exists:
                stmt += ln + "ORDER BY attr_order;"
            stmt = stmt.format(schema=self.schema, table=table_name)

            cur = self.conn.cursor()
            cur.execute(stmt)
            res = self._get_dict_result(cur)
        return res

    def get_min_max_info(self, table_name):
        result = {}
        # Get all 'c'heck constraints for this table
        if self.schema:
            constraints_cur = self.conn.cursor()
            # this query returns the clause check intervals that are similar to:
            #       ([numero_pisos]>=(1) AND [numero_pisos]<=(100))
            query = """
                SELECT CHECK_CLAUSE
                FROM
                    INFORMATION_SCHEMA.CHECK_CONSTRAINTS cc INNER JOIN
                    INFORMATION_SCHEMA.CONSTRAINT_COLUMN_USAGE c
                        ON cc.CONSTRAINT_NAME = c.CONSTRAINT_NAME
                        AND cc.CONSTRAINT_SCHEMA = c.CONSTRAINT_SCHEMA
                WHERE
                    cc.CONSTRAINT_SCHEMA = '{schema}'
                    AND TABLE_NAME = '{table}'
                """.format(
                schema=self.schema, table=table_name
            )

            constraints_cur.execute(query)

            # Create a mapping in the form of
            #
            # fieldname: (min, max)
            constraint_mapping = dict()
            for constraint in constraints_cur:
                # The regex takes the query results (e.g. '([numero_pisos]>=(1) AND [numero_pisos]<=(100))')
                # and gets the field name (regex-group 1), the minimum value (regex-group 2),
                # and the maximum value (regex-group 4) of the field for each register
                m = re.match(
                    r"\(\[(.*)\]>=\(([+-]?[0-9]+(?:\.[0-9]+)?(?:[eE][+-]?[0-9]+)?)\) AND \[(.*)\]<=\(([+-]?[0-9]+(?:\.[0-9]+)?(?:[eE][+-]?[0-9]+)?)\)\)",
                    constraint[0],
                )

                if m:
                    constraint_mapping[m.group(1)] = (m.group(2), m.group(4))

            result = constraint_mapping

        return result

    def get_relations_info(self, filter_layer_list=[]):
        result = []

        if self.schema:
            cur = self.conn.cursor()
            schema_where1 = (
                "AND KCU1.CONSTRAINT_SCHEMA = '{}'".format(self.schema)
                if self.schema
                else ""
            )
            schema_where2 = (
                "AND KCU2.CONSTRAINT_SCHEMA = '{}'".format(self.schema)
                if self.schema
                else ""
            )
            filter_layer_where = ""
            if filter_layer_list:
                filter_layer_where = "AND KCU1.TABLE_NAME IN ('{}')".format(
                    "','".join(filter_layer_list)
                )

            strength_field = ""
            strength_join = ""
            cardinality_max_field = ""
            cardinality_max_join = ""
            if self._table_exists(METAATTRS_TABLE):
                strength_field = ",META_ATTRS.attr_value as strength"
                strength_join = """
                LEFT JOIN {schema}.t_ili2db_attrname AS ATTRNAME
                    ON ATTRNAME.sqlname = KCU1.COLUMN_NAME AND ATTRNAME.{colowner} = KCU1.TABLE_NAME AND ATTRNAME.target = KCU2.TABLE_NAME
                LEFT JOIN {schema}.t_ili2db_meta_attrs AS META_ATTRS
                    ON META_ATTRS.ilielement = ATTRNAME.iliname AND META_ATTRS.attr_name = 'ili2db.ili.assocKind'
                    """.format(
                    schema=self.schema,
                    colowner="owner" if self.ili_version() == 3 else "colowner",
                )

                cardinality_max_field = (
                    ",META_ATTRS_CARDINALITY.attr_value as cardinality_max"
                )
                cardinality_max_join = """
                LEFT JOIN {schema}.t_ili2db_attrname AS ATTRNAME_CARDINALITY
                    ON ATTRNAME_CARDINALITY.sqlname = KCU1.COLUMN_NAME AND ATTRNAME_CARDINALITY.{colowner} = KCU1.TABLE_NAME AND ATTRNAME_CARDINALITY.target = KCU2.TABLE_NAME
                LEFT JOIN {schema}.t_ili2db_meta_attrs AS META_ATTRS_CARDINALITY
                    ON META_ATTRS_CARDINALITY.ilielement = ATTRNAME_CARDINALITY.iliname AND META_ATTRS_CARDINALITY.attr_name = 'ili2db.ili.attrCardinalityMax'
                    """.format(
                    schema=self.schema,
                    colowner="owner" if self.ili_version() == 3 else "colowner",
                )

            query = """
                SELECT
                    KCU1.CONSTRAINT_NAME AS constraint_name
                    ,KCU1.TABLE_NAME AS referencing_table
                    ,KCU1.COLUMN_NAME AS referencing_column
                    ,KCU2.CONSTRAINT_SCHEMA AS constraint_schema
                    ,KCU2.TABLE_NAME AS referenced_table
                    ,KCU2.COLUMN_NAME AS referenced_column
                    ,KCU1.ORDINAL_POSITION AS ordinal_position
                    {strength_field}
                    {cardinality_max_field}
                FROM INFORMATION_SCHEMA.REFERENTIAL_CONSTRAINTS AS RC

                INNER JOIN INFORMATION_SCHEMA.KEY_COLUMN_USAGE AS KCU1
                    ON KCU1.CONSTRAINT_CATALOG = RC.CONSTRAINT_CATALOG
                    AND KCU1.CONSTRAINT_SCHEMA = RC.CONSTRAINT_SCHEMA
                    AND KCU1.CONSTRAINT_NAME = RC.CONSTRAINT_NAME

                INNER JOIN INFORMATION_SCHEMA.KEY_COLUMN_USAGE AS KCU2
                    ON KCU2.CONSTRAINT_CATALOG = RC.UNIQUE_CONSTRAINT_CATALOG
                    AND KCU2.CONSTRAINT_SCHEMA = RC.UNIQUE_CONSTRAINT_SCHEMA
                    AND KCU2.CONSTRAINT_NAME = RC.UNIQUE_CONSTRAINT_NAME
                    AND KCU2.ORDINAL_POSITION = KCU1.ORDINAL_POSITION
                {strength_join}
                {cardinality_max_join}

                WHERE 1=1 {schema_where1} {schema_where2} {filter_layer_where}
                order by constraint_name, ordinal_position
                """.format(
                schema_where1=schema_where1,
                schema_where2=schema_where2,
                filter_layer_where=filter_layer_where,
                strength_field=strength_field,
                strength_join=strength_join,
                cardinality_max_field=cardinality_max_field,
                cardinality_max_join=cardinality_max_join,
            )
            cur.execute(query)
            result = self._get_dict_result(cur)

        return result

    def get_iliname_dbname_mapping(self, sqlnames=list()):
        """Note: the parameter sqlnames is only used for ili2db version 3 relation creation"""
        result = {}
        # Map domain ili name with its correspondent mssql name
        if self.schema and self.metadata_exists():
            cur = self.conn.cursor()

            where = ""
            if sqlnames:
                names = "'" + "','".join(sqlnames) + "'"
                where = "WHERE sqlname IN ({})".format(names)

            cur.execute(
                """SELECT iliname, sqlname
                           FROM {schema}.t_ili2db_classname
                           {where}
                        """.format(
                    schema=self.schema, where=where
                )
            )

            result = self._get_dict_result(cur)

        return result

    def get_models(self):
        """Needed for exportmodels"""
        result = {}
        # Get MODELS
        if self.schema:
            cur = self.conn.cursor()

            cur.execute(
                """SELECT modelname, content
                           FROM {schema}.t_ili2db_model
                        """.format(
                    schema=self.schema
                )
            )
            result = self._get_dict_result(cur)

        return result

    def get_classili_classdb_mapping(self, models_info, extended_classes):
        result = {}
        if self.schema:
            cur = self.conn.cursor()
            class_names = (
                "'"
                + "','".join(list(models_info.keys()) + list(extended_classes.keys()))
                + "'"
            )
            cur.execute(
                """SELECT iliname, sqlname
                           FROM {schema}.t_ili2db_classname
                           WHERE iliname IN ({class_names})
                        """.format(
                    schema=self.schema, class_names=class_names
                )
            )
            result = self._get_dict_result(cur)
        return result

    def get_attrili_attrdb_mapping(self, attrs_list):
        result = {}
        if self.schema:
            cur = self.conn.cursor()
            attr_names = "'" + "','".join(attrs_list) + "'"

            cur.execute(
                """SELECT iliname, sqlname, owner
                           FROM {schema}.t_ili2db_attrname
                           WHERE iliname IN ({attr_names})
                        """.format(
                    schema=self.schema, attr_names=attr_names
                )
            )
            result = self._get_dict_result(cur)
        return result

    def get_attrili_attrdb_mapping_by_owner(self, owners):
        result = {}
        if self.schema:
            cur = self.conn.cursor()
            owner_names = "'" + "','".join(owners) + "'"
            cur.execute(
                """SELECT iliname, sqlname, owner
                           FROM {schema}.t_ili2db_attrname
                           WHERE owner IN ({owner_names})
                        """.format(
                    schema=self.schema, owner_names=owner_names
                )
            )
            result = self._get_dict_result(cur)
        return result

    def _get_dict_result(self, cur):
        columns = [column[0] for column in cur.description]

        res = []
        for row in cur.fetchall():
            my_rec = dict(zip(columns, row))
            res.append(my_rec)

        return res

    def ili_version(self):
        cur = self.conn.cursor()
        cur.execute(
            """SELECT count(COLUMN_NAME)
FROM INFORMATION_SCHEMA.COLUMNS
WHERE TABLE_SCHEMA='{schema}'
	AND(TABLE_NAME='t_ili2db_attrname' OR TABLE_NAME='t_ili2db_model')
                       AND(COLUMN_NAME='owner' OR COLUMN_NAME='file')""".format(
                schema=self.schema
            )
        )

        res = cur.fetchone()[0]
        print(res)
        if res > 0:
            self.new_message.emit(
                Qgis.Warning,
                "DB schema created with ili2db version 3. Better use version 4.",
            )
            return 3
        else:
            return 4

    def get_basket_handling(self):
        if self.schema and self._table_exists(SETTINGS_TABLE):
            cur = self.conn.cursor()
            cur.execute(
                """SELECT setting
                            FROM {schema}.{table}
                            WHERE tag = ?
                            """.format(
                    schema=self.schema, table=SETTINGS_TABLE
                ),
                ("ch.ehi.ili2db.BasketHandling",),
            )
            content = cur.fetchone()
            if content:
                return content[0] == "readWrite"
        return False

    def get_baskets_info(self):
        result = {}
        if self.schema and self._table_exists(BASKET_TABLE):
            cur = self.conn.cursor()
            cur.execute(
                """SELECT b.t_id as basket_t_id,
                            b.t_ili_tid as basket_t_ili_tid,
                            b.topic as topic,
                            d.t_id as dataset_t_id,
                            d.datasetname as datasetname from {schema}.{basket_table} b
                            JOIN {schema}.{dataset_table} d
                            ON b.dataset = d.t_id
                        """.format(
                    schema=self.schema,
                    basket_table=BASKET_TABLE,
                    dataset_table=DATASET_TABLE,
                )
            )
            result = self._get_dict_result(cur)
        return result

    def get_datasets_info(self):
        result = {}
        if self.schema and self._table_exists(DATASET_TABLE):
            cur = self.conn.cursor()
            cur.execute(
                """SELECT t_id, datasetname
                           FROM {schema}.{dataset_table}
                        """.format(
                    schema=self.schema, dataset_table=DATASET_TABLE
                )
            )
            result = self._get_dict_result(cur)
        return result

    def create_dataset(self, datasetname):
        if self.schema and self._table_exists(DATASET_TABLE):
            cur = self.conn.cursor()
            try:
                cur.execute(
                    """
                    INSERT INTO {schema}.{dataset_table} VALUES (NEXT VALUE FOR {schema}.{sequence}, ?)
                    """.format(
                        schema=self.schema,
                        sequence="t_ili2db_seq",
                        dataset_table=DATASET_TABLE,
                    ),
                    datasetname,
                )
                self.conn.commit()
                return True, self.tr('Successfully created dataset "{}".').format(
                    datasetname
                )
            except pyodbc.errors.UniqueViolation as e:
                return False, self.tr('Dataset with name "{}" already exists.').format(
                    datasetname
                )
        return False, self.tr('Could not create dataset "{}".').format(datasetname)

    def rename_dataset(self, tid, datasetname):
        if self.schema and self._table_exists(DATASET_TABLE):
            cur = self.conn.cursor()
            try:
                cur.execute(
                    """
                    UPDATE {schema}.{dataset_table} SET datasetname = ? WHERE {tid_name} = {tid}
                    """.format(
                        schema=self.schema,
                        dataset_table=DATASET_TABLE,
                        tid_name=self.tid,
                        tid=tid,
                    ),
                    datasetname,
                )
                self.conn.commit()
                return True, self.tr('Successfully created dataset "{}".').format(
                    datasetname
                )
            except pyodbc.errors.UniqueViolation as e:
                return False, self.tr('Dataset with name "{}" already exists.').format(
                    datasetname
                )
        return False, self.tr('Could not create dataset "{}".').format(datasetname)

    def get_topics_info(self):
        result = {}
        if self.schema and self._table_exists("t_ili2db_classname"):
            cur = self.conn.cursor()
            cur.execute(
                """
                    SELECT DISTINCT PARSENAME(cn.iliname,1) as model,
                    PARSENAME(cn.iliname,2) as topic
                    FROM {schema}.t_ili2db_classname as cn
                    JOIN {schema}.t_ili2db_table_prop as tp
                    ON cn.sqlname = tp.tablename
					WHERE PARSENAME(cn.iliname,3) != '' and tp.setting != 'ENUM'
                """.format(
                    schema=self.schema
                )
            )
            result = self._get_dict_result(cur)
        return result

    def create_basket(self, dataset_tid, topic):
        if self.schema and self._table_exists(BASKET_TABLE):
            cur = self.conn.cursor()
            cur.execute(
                """
                    SELECT * FROM {schema}.{basket_table}
                    WHERE dataset = {dataset_tid} and topic = '{topic}'
                """.format(
                    schema=self.schema,
                    basket_table=BASKET_TABLE,
                    dataset_tid=dataset_tid,
                    topic=topic,
                )
            )
            if cur.fetchone():
                return False, self.tr('Basket for topic "{}" already exists.').format(
                    topic
                )
            try:
                cur.execute(
                    """
                    INSERT INTO {schema}.{basket_table} ({tid_name}, dataset, topic, {tilitid_name}, attachmentkey )
                    VALUES (NEXT VALUE FOR {schema}.{sequence}, {dataset_tid}, '{topic}', NEWID(), 'Qgis Model Baker')
                """.format(
                        schema=self.schema,
                        sequence="t_ili2db_seq",
                        tid_name=self.tid,
                        tilitid_name=self.tilitid,
                        basket_table=BASKET_TABLE,
                        dataset_tid=dataset_tid,
                        topic=topic,
                    )
                )
                self.conn.commit()
                return True, self.tr(
                    'Successfully created basket for topic "{}".'
                ).format(topic)
            except pyodbc.errors.Error as e:
                error_message = " ".join(e.args)
                return False, self.tr(
                    'Could not create basket for topic "{}": {}'
                ).format(topic, error_message)
        return False, self.tr('Could not create basket for topic "{}".').format(topic)

    def get_tid_handling(self):
        if self.schema and self._table_exists(SETTINGS_TABLE):
            cur = self.conn.cursor()
            cur.execute(
                """SELECT setting
                            FROM {schema}.{table}
                            WHERE tag = ?
                            """.format(
                    schema=self.schema, table=SETTINGS_TABLE
                ),
                ("ch.ehi.ili2db.TidHandling",),
            )
            content = cur.fetchone()
            if content:
                return content[0] == "property"
        return False

    def get_ili2db_settings(self):
        result = {}
        if self._table_exists(SETTINGS_TABLE):
            cur = self.conn.cursor()
            cur.execute(
                """SELECT tag, setting
                            FROM {schema}.{table}
                            """.format(
                    schema=self.schema, table=SETTINGS_TABLE
                )
            )
            result = self._get_dict_result(cur)
        return result



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 2022-07-17
        git sha              : :%H$
        copyright            : (C) 2022 by Dave Signer
        email                : david at opengis ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import configparser
import os
from typing import Union

from ..libs.toppingmaker import ProjectTopping, Target
from ..libs.toppingmaker.utils import slugify
from .ili2dbsettings import Ili2dbSettings


class MetaConfig(object):
    """
    Class containing all the data needed to create the metaconfiguration file.
    Having the section [CONFIGURATION] containing the links to the referencedata nd the projecttoppingfile.
    And the section [ch.ehi.ili2db] containing all the ili2db settings.
    """

    METACONFIG_TYPE = "metaconfig"
    REFERENCEDATA_TYPE = "referencedata"
    METAATTR_TYPE = "metaattributes"
    SQLSCRIPT_TYPE = "sql"

    def __init__(self):
        self.referencedata_paths = []
        self.projecttopping_path = None
        self.ili2db_settings = Ili2dbSettings()

    def update_referencedata_paths(self, value: Union[list, bool]):
        if isinstance(value, str):
            value = [value]
        self.referencedata_paths.extend(value)

    def update_projecttopping_path(self, value: str):
        self.projecttopping_path = value

    def generate_files(self, target: Target):
        """
        Example content:

        [CONFIGURATION]
        qgis.modelbaker.projecttopping=ilidata:ch.opengis.config.KbS_LV95_V1_4_projecttopping
        ch.interlis.referenceData=ilidata:ch.opengis.config.KbS_Codetexte_V1_4

        [ch.ehi.ili2db]
        defaultSrsCode = 2056
        smart2Inheritance = true
        strokeArcs = false
        importTid = true
        createTidCol = false
        models = KbS_Basis_V1_4
        preScript=ilidata:ch.opengis.config.KbS_LV95_V1_4_prescript
        iliMetaAttrs=ilidata:ch.opengis.config.KbS_LV95_V1_4_toml
        """

        configuration_section = {}

        # append project topping and reference data links
        if self.projecttopping_path:
            # generate toppingfile of projettopping (most possibly already an id, so no generation needed)
            projecttopping_link = self._generate_toppingfile_link(
                target, ProjectTopping.PROJECTTOPPING_TYPE, self.projecttopping_path
            )
            configuration_section[
                "qgis.modelbaker.projecttopping"
            ] = projecttopping_link

        if self.referencedata_paths:
            # generate toppingfiles of the reference data
            referencedata_links = ",".join(
                [
                    self._generate_toppingfile_link(
                        target, MetaConfig.REFERENCEDATA_TYPE, path
                    )
                    for path in self.referencedata_paths
                ]
            )
            configuration_section["ch.interlis.referenceData"] = referencedata_links

        ili2db_section = {}

        # append models and the ili2db parameters
        ili2db_section["models"] = ",".join(self.ili2db_settings.models)
        ili2db_section.update(self.ili2db_settings.parameters)

        # generate metaattr and prescript / postscript files
        if self.ili2db_settings.metaattr_path:
            ili2db_section["iliMetaAttrs"] = self._generate_toppingfile_link(
                target, MetaConfig.METAATTR_TYPE, self.ili2db_settings.metaattr_path
            )
        if self.ili2db_settings.prescript_path:
            ili2db_section["preScript"] = self._generate_toppingfile_link(
                target, MetaConfig.SQLSCRIPT_TYPE, self.ili2db_settings.prescript_path
            )
        if self.ili2db_settings.postscript_path:
            ili2db_section["postScript"] = self._generate_toppingfile_link(
                target, MetaConfig.SQLSCRIPT_TYPE, self.ili2db_settings.postscript_path
            )

        # make the full configuration
        metaconfig = configparser.ConfigParser()
        # write the keys as they are (don't make em lowercase)
        metaconfig.optionxform = str
        metaconfig["CONFIGURATION"] = configuration_section
        metaconfig["ch.ehi.ili2db"] = ili2db_section

        # write INI file
        metaconfig_slug = f"{slugify(target.projectname)}.ini"
        absolute_filedir_path, relative_filedir_path = target.filedir_path(
            MetaConfig.METACONFIG_TYPE
        )

        with open(
            os.path.join(absolute_filedir_path, metaconfig_slug), "w"
        ) as configfile:
            metaconfig.write(configfile)

        return target.path_resolver(target, metaconfig_slug, MetaConfig.METACONFIG_TYPE)

    def _generate_toppingfile_link(self, target: Target, type, path):
        if not os.path.isfile(path):
            # it's already an id pointing to somewhere, no toppingfile needs to be created
            return path
        # copy file from path to our target, and return the ilidata_pathresolved id
        return target.toppingfile_link(type, path)



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 15/06/17
        git sha              : :%H$
        copyright            : (C) 2017 by OPENGIS.ch
        email                : info@opengis.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
import glob
import logging
import os
import re
import urllib.parse
import xml.etree.ElementTree as ET
from enum import Enum

from PyQt5.QtCore import QObject, Qt, pyqtSignal
from PyQt5.QtGui import QPalette, QRegion, QStandardItem, QStandardItemModel
from PyQt5.QtWidgets import QGridLayout, QItemDelegate, QLabel, QStyle, QWidget
from qgis.core import Qgis, QgsMessageLog

from ..utils.qt_utils import download_file
from .ili2dbutils import get_all_modeldir_in_path


class IliCache(QObject):
    ns = {"ili23": "http://www.interlis.ch/INTERLIS2.3"}

    new_message = pyqtSignal(int, str)

    def __init__(self, configuration, single_ili_file=None):
        QObject.__init__(self)
        self.cache_path = os.path.expanduser("~/.ilicache")
        self.information_file = "ilimodels.xml"
        self.repositories = dict()
        self.base_configuration = configuration
        self.single_ili_file = single_ili_file
        self.model = IliModelItemModel()
        self.directories = None
        if self.base_configuration:
            self.directories = self.base_configuration.model_directories

    def refresh(self):
        if not self.directories is None:
            for directory in self.directories:
                self.process_model_directory(directory)

        if not self.single_ili_file is None:
            if os.path.exists(self.single_ili_file):
                self.process_single_ili_file()

    def process_model_directory(self, path):
        if path[0] == "%":
            pass
        else:
            # download remote and local repositories
            self.download_repository(path)

            if os.path.isdir(path):
                # additional recursive search of paths containing ili files (without ilimodel.xml)
                get_all_modeldir_in_path(path, self.process_local_ili_folder)

    def process_single_ili_file(self):
        models = self.process_ili_file(self.single_ili_file)
        self.repositories["no_repo"] = sorted(
            models, key=lambda m: m["version"], reverse=True
        )
        self.model.set_repositories(self.repositories)

    def file_url(self, url, file):
        if url is None:
            return file
        elif os.path.isdir(url):
            return os.path.join(url, file)
        else:
            return urllib.parse.urljoin(url + "/", file)

    def file_path(self, netloc, url, file):
        if url is None:
            return file
        elif os.path.isdir(url):
            return os.path.join(url, file)
        else:
            netloc = "" if netloc is None else netloc
            return os.path.join(self.cache_path, netloc, file)

    def download_repository(self, url):
        """
        Downloads the informationfile (default: ilimodels.xml) and ilisite.xml files from the provided url
        and updates the local cache.
        """
        netloc = urllib.parse.urlsplit(url)[1] if not os.path.isdir(url) else url

        information_file_url = self.file_url(url, self.information_file)
        ilisite_url = self.file_url(url, "ilisite.xml")
        logger = logging.getLogger(__name__)

        if os.path.isdir(url):
            # continue with the local file
            if os.path.exists(information_file_url):
                self._process_informationfile(information_file_url, netloc, url)
            else:
                logger.warning(
                    self.tr("Could not find local file {}").format(information_file_url)
                )
            if os.path.exists(ilisite_url):
                self._process_ilisite(ilisite_url)
            else:
                logger.warning(
                    self.tr("Could not find local file  {}").format(ilisite_url)
                )
        else:
            netloc_dir = os.path.join(self.cache_path, netloc)
            os.makedirs(netloc_dir, exist_ok=True)
            information_file_path = os.path.join(netloc_dir, self.information_file)
            ilisite_path = os.path.join(self.cache_path, netloc, "ilisite.xml")

            # download ilimodels.xml
            download_file(
                information_file_url,
                information_file_path,
                on_success=lambda: self._process_informationfile(
                    information_file_path, netloc, url
                ),
                on_error=lambda error, error_string: logger.warning(
                    self.tr("Could not download {url} ({message})").format(
                        url=information_file_url, message=error_string
                    )
                ),
            )

            # download ilisite.xml
            download_file(
                ilisite_url,
                ilisite_path,
                on_success=lambda: self._process_ilisite(ilisite_path),
                on_error=lambda error, error_string: logger.warning(
                    self.tr("Could not download {url} ({message})").format(
                        url=ilisite_url, message=error_string
                    )
                ),
            )

    def _process_ilisite(self, file):
        """
        Parses the ilisite.xml provided in ``file`` and recursively downloads any subidiary sites.
        """
        try:
            root = ET.parse(file).getroot()
        except ET.ParseError as e:
            QgsMessageLog.logMessage(
                self.tr(
                    "Could not parse ilisite file `{file}` ({exception})".format(
                        file=file, exception=str(e)
                    )
                ),
                self.tr("QGIS Model Baker"),
            )
            return

        for site in root.iter(
            "{http://www.interlis.ch/INTERLIS2.3}IliSite09.SiteMetadata.Site"
        ):
            subsite = site.find("ili23:subsidiarySite", self.ns)
            if subsite is not None:
                for location in subsite.findall(
                    "ili23:IliSite09.RepositoryLocation_", self.ns
                ):
                    value = self.get_element_text(location.find("ili23:value", self.ns))
                    if value:
                        self.download_repository(value)

    def _process_informationfile(self, file, netloc, url):
        """
        Parses ilimodels.xml provided in ``file`` and updates the local repositories cache.
        """

        try:
            root = ET.parse(file).getroot()
        except ET.ParseError as e:
            QgsMessageLog.logMessage(
                self.tr(
                    "Could not parse ilimodels file `{file}` ({exception})".format(
                        file=file, exception=str(e)
                    )
                ),
                self.tr("QGIS Model Baker"),
            )
            return

        self.repositories[netloc] = list()
        repo_models = list()
        for repo in root.iter(
            "{http://www.interlis.ch/INTERLIS2.3}IliRepository09.RepositoryIndex"
        ):
            for model_metadata in repo.findall(
                "ili23:IliRepository09.RepositoryIndex.ModelMetadata", self.ns
            ):
                model = dict()
                model["name"] = self.get_element_text(
                    element=model_metadata.find("ili23:Name", self.ns)
                )
                if model["name"]:
                    model["version"] = self.get_element_text(
                        model_metadata.find("ili23:Version", self.ns)
                    )
                    model["repository"] = netloc
                    repo_models.append(model)

        for repo in root.iter(
            "{http://www.interlis.ch/INTERLIS2.3}IliRepository20.RepositoryIndex"
        ):
            for model_metadata in repo.findall(
                "ili23:IliRepository20.RepositoryIndex.ModelMetadata", self.ns
            ):
                model = dict()
                model["name"] = self.get_element_text(
                    model_metadata.find("ili23:Name", self.ns)
                )
                if model["name"]:
                    model["version"] = self.get_element_text(
                        model_metadata.find("ili23:Version", self.ns)
                    )
                    model["repository"] = netloc
                    repo_models.append(model)

        self.repositories[netloc] = sorted(
            repo_models,
            key=lambda m: m["version"] if m["version"] else "0",
            reverse=True,
        )

        self.model.set_repositories(self.repositories)

    def process_local_ili_folder(self, path):
        """
        Parses all .ili files in the given ``path`` (non-recursively)
        """
        models = list()
        fileModels = list()
        for ilifile in glob.iglob(os.path.join(path, "*.ili")):
            fileModels = self.process_ili_file(ilifile)
            models.extend(fileModels)

        self.repositories[path] = sorted(
            models, key=lambda m: m["version"], reverse=True
        )

        self.model.set_repositories(self.repositories)

    def process_ili_file(self, ilifile):
        fileModels = list()
        try:
            fileModels = self.parse_ili_file(ilifile, "utf-8")
        except UnicodeDecodeError:
            try:
                fileModels = self.parse_ili_file(ilifile, "latin1")
                self.new_message.emit(
                    Qgis.Warning,
                    self.tr(
                        "Even though the ili file `{}` could be read, it is not in UTF-8. Please encode your ili models in UTF-8.".format(
                            os.path.basename(ilifile)
                        )
                    ),
                )
            except UnicodeDecodeError as e:
                self.new_message.emit(
                    Qgis.Critical,
                    self.tr(
                        "Could not parse ili file `{}` with UTF-8 nor Latin-1 encodings. Please encode your ili models in UTF-8.".format(
                            os.path.basename(ilifile)
                        )
                    ),
                )
                QgsMessageLog.logMessage(
                    self.tr(
                        "Could not parse ili file `{ilifile}`. We suggest you to encode it in UTF-8. ({exception})".format(
                            ilifile=ilifile, exception=str(e)
                        )
                    ),
                    self.tr("QGIS Model Baker"),
                )
                fileModels = list()

        return fileModels

    def parse_ili_file(self, ilipath, encoding):
        """
        Parses an ili file returning models and version data
        """
        models = list()
        re_model = re.compile(r"\s*MODEL\s*([\w\d_-]+).*")
        re_model_version = re.compile(r'VERSION "([ \w\d\._-]+)".*')
        with open(ilipath, "r", encoding=encoding) as file:
            model = None
            for lineno, line in enumerate(file):
                line = line.split("!!")[0]
                result = re_model.search(line)
                if result:
                    model = dict()
                    model["name"] = result.group(1)
                    model["version"] = ""
                    model["repository"] = ilipath
                    models += [model]

                result = re_model_version.search(line)
                if result:
                    if not model:
                        raise RuntimeError(
                            "VERSION tag found in file {}:{} without previous MODEL definition.".format(
                                ilipath, lineno
                            )
                        )
                    model["version"] = result.group(1)
                    model = None

        return models

    @property
    def model_names(self):
        names = list()
        for repo in self.repositories.values():
            for model in repo:
                names.append(model["name"])
        return names

    def get_element_text(self, element):
        if element is not None:
            return element.text
        return None


class IliModelItemModel(QStandardItemModel):
    class Roles(Enum):
        ILIREPO = Qt.UserRole + 1
        VERSION = Qt.UserRole + 2

        def __int__(self):
            return self.value

    def __init__(self, parent=None):
        super().__init__(0, 1, parent)

    def set_repositories(self, repositories):
        self.clear()
        row = 0
        names = list()

        for repository in repositories.values():

            for model in repository:
                # in case there is more than one version of the model with the same name, it shouldn't load it twice
                if any(model["name"] == s for s in names):
                    continue

                item = QStandardItem()
                item.setData(model["name"], int(Qt.DisplayRole))
                item.setData(model["name"], int(Qt.EditRole))
                item.setData(model["repository"], int(IliModelItemModel.Roles.ILIREPO))
                item.setData(model["version"], int(IliModelItemModel.Roles.VERSION))

                names.append(model["name"])
                self.appendRow(item)
                row += 1


class ModelCompleterDelegate(QItemDelegate):
    """
    A item delegate for the autocompleter of model dialogs.
    It shows the source repository next to the model name.
    """

    def __init__(self):
        super().__init__()
        self.widget = QWidget()
        self.widget.setLayout(QGridLayout())
        self.widget.layout().setContentsMargins(2, 0, 0, 0)
        self.model_label = QLabel()
        self.model_label.setAttribute(Qt.WA_TranslucentBackground)
        self.repository_label = QLabel()
        self.repository_label.setAlignment(Qt.AlignRight)
        self.widget.layout().addWidget(self.model_label, 0, 0)
        self.widget.layout().addWidget(self.repository_label, 0, 1)

    def paint(self, painter, option, index):
        option.index = index
        super().paint(painter, option, index)

    def drawDisplay(self, painter, option, rect, text):
        repository = option.index.data(int(IliModelItemModel.Roles.ILIREPO))
        option.index.data(int(IliModelItemModel.Roles.VERSION))
        self.repository_label.setText(
            '<font color="#666666"><i>{repository}</i></font>'.format(
                repository=repository
            )
        )
        self.model_label.setText("{model}".format(model=text))
        self.widget.setMinimumSize(rect.size())

        model_palette = option.palette
        if option.state & QStyle.State_Selected:
            model_palette.setColor(
                QPalette.WindowText,
                model_palette.color(QPalette.Active, QPalette.HighlightedText),
            )

        self.widget.render(painter, rect.topLeft(), QRegion(), QWidget.DrawChildren)


class IliDataCache(IliCache):

    file_download_succeeded = pyqtSignal(str, str)
    file_download_failed = pyqtSignal(str, str)
    model_refreshed = pyqtSignal(int)

    def __init__(self, configuration, type="metaconfig", models=None):
        IliCache.__init__(self, configuration)
        self.cache_path = os.path.expanduser("~/.ilimetaconfigcache")
        self.information_file = "ilidata.xml"

        self.model = IliDataItemModel()
        self.model.modelReset.connect(
            lambda: self.model_refreshed.emit(self.model.rowCount())
        )

        self.filter_models = models.split(";") if models else []
        self.type = type
        self.directories = (
            self.base_configuration.metaconfig_directories
            if self.base_configuration
            else []
        )

    def process_model_directory(self, path):
        # download remote and local repositories
        self.download_repository(path)

    def _process_informationfile(self, file, netloc, url):
        """
        Parses ilidata.xml provided in ``file`` and updates the local repositories cache.
        """
        try:
            root = ET.parse(file).getroot()
        except ET.ParseError as e:
            QgsMessageLog.logMessage(
                self.tr(
                    "Could not parse ilidata file `{file}` ({exception})".format(
                        file=file, exception=str(e)
                    )
                ),
                self.tr("QGIS Model Baker"),
            )
            return

        model_code_regex = re.compile("http://codes.interlis.ch/model/(.*)")
        type_code_regex = re.compile("http://codes.interlis.ch/type/(.*)")
        tool_code_regex = re.compile("http://codes.opengis.ch/(.*)")

        self.repositories[netloc] = list()
        repo_metaconfigs = list()
        for repo in root.iter(
            "{http://www.interlis.ch/INTERLIS2.3}DatasetIdx16.DataIndex"
        ):
            for metaconfig_metadata in repo.findall(
                "ili23:DatasetIdx16.DataIndex.DatasetMetadata", self.ns
            ):
                categories_element = metaconfig_metadata.find(
                    "ili23:categories", self.ns
                )
                if categories_element is not None:
                    models = []
                    type = ""
                    for category in categories_element.findall(
                        "ili23:DatasetIdx16.Code_", self.ns
                    ):
                        category_value = self.get_element_text(
                            category.find("ili23:value", self.ns)
                        )
                        if category_value:
                            if model_code_regex.search(category_value):
                                models.append(
                                    model_code_regex.search(category_value).group(1)
                                )
                            if type_code_regex.search(category_value):
                                type = type_code_regex.search(category_value).group(1)
                            if tool_code_regex.search(category_value):
                                tool_code_regex.search(category_value).group(1)
                    if (
                        not any(model in models for model in self.filter_models)
                        or type != self.type
                    ):
                        continue
                    title = list()
                    for title_element in metaconfig_metadata.findall(
                        "ili23:title", self.ns
                    ):
                        for multilingual_text_element in title_element.findall(
                            "ili23:DatasetIdx16.MultilingualText", self.ns
                        ):
                            for (
                                localised_text_element
                            ) in multilingual_text_element.findall(
                                "ili23:LocalisedText", self.ns
                            ):
                                for (
                                    localised_text_element
                                ) in localised_text_element.findall(
                                    "ili23:DatasetIdx16.LocalisedText", self.ns
                                ):
                                    title_information = {
                                        "language": self.get_element_text(
                                            localised_text_element.find(
                                                "ili23:Language", self.ns
                                            )
                                        ),
                                        "text": self.get_element_text(
                                            localised_text_element.find(
                                                "ili23:Text", self.ns
                                            )
                                        ),
                                    }
                                    title.append(title_information)

                    model_link_names = []
                    for basket_element in metaconfig_metadata.findall(
                        "ili23:baskets", self.ns
                    ):
                        for basket_metadata in basket_element.findall(
                            "ili23:DatasetIdx16.DataIndex.BasketMetadata", self.ns
                        ):
                            for model_element in basket_metadata.findall(
                                "ili23:model", self.ns
                            ):
                                for model_link in model_element.findall(
                                    "ili23:DatasetIdx16.ModelLink", self.ns
                                ):
                                    model_link_names.append(
                                        self.get_element_text(
                                            model_link.find("ili23:name", self.ns)
                                        )
                                    )

                    for files_element in metaconfig_metadata.findall(
                        "ili23:files", self.ns
                    ):
                        for data_file in files_element.findall(
                            "ili23:DatasetIdx16.DataFile", self.ns
                        ):
                            for file_element in data_file.findall(
                                "ili23:file", self.ns
                            ):
                                for file in file_element.findall(
                                    "ili23:DatasetIdx16.File", self.ns
                                ):
                                    path = self.get_element_text(
                                        file.find("ili23:path", self.ns)
                                    )

                                    metaconfig = dict()
                                    metaconfig["id"] = self.get_element_text(
                                        metaconfig_metadata.find("ili23:id", self.ns)
                                    )

                                    metaconfig["version"] = self.get_element_text(
                                        metaconfig_metadata.find(
                                            "ili23:version", self.ns
                                        )
                                    )
                                    metaconfig["owner"] = self.get_element_text(
                                        metaconfig_metadata.find("ili23:owner", self.ns)
                                    )

                                    metaconfig["repository"] = netloc
                                    metaconfig["url"] = url
                                    metaconfig["models"] = models
                                    metaconfig["relative_file_path"] = path
                                    if title is not None:
                                        metaconfig["title"] = title
                                    else:
                                        metaconfig["title"] = None

                                    metaconfig["model_link_names"] = model_link_names
                                    repo_metaconfigs.append(metaconfig)

        self.repositories[netloc] = sorted(
            repo_metaconfigs,
            key=lambda m: m["version"] if m["version"] else "0",
            reverse=True,
        )

        self.model.set_repositories(self.repositories)

    def download_file(self, netloc, url, file, dataset_id=None):
        """
        Downloads the given file from the given url to the local cache.
        passes the local file path or the id (for information) to signals.
        Returns the file path immediately (might not be downloaded yet)
        """
        file_url = self.file_url(url, file)

        if url is None or os.path.isdir(url):
            file_path = file_url
            # continue with the local file
            if os.path.exists(file_url):
                self.file_download_succeeded.emit(dataset_id, file_url)
            else:
                self.file_download_failed.emit(
                    dataset_id,
                    self.tr("Could not find local file  {}").format(file_url),
                )
        else:
            file_path = os.path.join(self.cache_path, netloc, file)
            file_dir = os.path.dirname(file_path)
            os.makedirs(file_dir, exist_ok=True)

            download_file(
                file_url,
                file_path,
                on_success=lambda: self.file_download_succeeded.emit(
                    dataset_id, file_path
                ),
                on_error=lambda error, error_string: self.file_download_failed.emit(
                    dataset_id,
                    self.tr("Could not download file {url} ({message})").format(
                        url=file_url, message=error_string
                    ),
                ),
            )
        return file_path


class IliDataItemModel(QStandardItemModel):
    class Roles(Enum):
        ILIREPO = Qt.UserRole + 1
        VERSION = Qt.UserRole + 2
        MODELS = Qt.UserRole + 3
        RELATIVEFILEPATH = Qt.UserRole + 4
        OWNER = Qt.UserRole + 5
        TITLE = Qt.UserRole + 6
        ID = Qt.UserRole + 7
        URL = Qt.UserRole + 8
        MODEL_LINKS = Qt.UserRole + 9

        def __int__(self):
            return self.value

    def __init__(self, parent=None):
        super().__init__(0, 1, parent)

    def set_repositories(self, repositories):
        self.beginResetModel()
        self.clear()
        row = 0
        ids = list()

        for repository in repositories.values():

            for dataitem in repository:
                if any(dataitem["id"] == s for s in ids):
                    continue

                item = QStandardItem()
                display_value = dataitem["id"]
                if dataitem["title"] and "text" in dataitem["title"][0]:
                    # since there is no multilanguage handling we take the first entry
                    display_value = dataitem["title"][0]["text"]

                item.setData(display_value, int(Qt.DisplayRole))
                item.setData(display_value, int(Qt.EditRole))
                item.setData(dataitem["id"], int(IliDataItemModel.Roles.ID))
                item.setData(
                    dataitem["repository"], int(IliDataItemModel.Roles.ILIREPO)
                )
                item.setData(dataitem["version"], int(IliDataItemModel.Roles.VERSION))
                item.setData(dataitem["models"], int(IliDataItemModel.Roles.MODELS))
                item.setData(
                    dataitem["relative_file_path"],
                    int(IliDataItemModel.Roles.RELATIVEFILEPATH),
                )
                item.setData(dataitem["owner"], int(IliDataItemModel.Roles.OWNER))
                item.setData(dataitem["title"], int(IliDataItemModel.Roles.TITLE))
                item.setData(dataitem["url"], int(IliDataItemModel.Roles.URL))

                item.setData(
                    dataitem["model_link_names"],
                    int(IliDataItemModel.Roles.MODEL_LINKS),
                )

                ids.append(dataitem["id"])
                self.appendRow(item)
                row += 1
        self.endResetModel()


class MetaConfigCompleterDelegate(QItemDelegate):
    """
    A item delegate for the autocompleter of metaconfig / topping dialogs.
    It shows the source repository (including model) next to the metaconfig id and the owner.
    """

    def __init__(self):
        super().__init__()
        self.widget = QWidget()
        self.widget.setLayout(QGridLayout())
        self.widget.layout().setContentsMargins(2, 0, 0, 0)
        self.metaconfig_label = QLabel()
        self.metaconfig_label.setAttribute(Qt.WA_TranslucentBackground)
        self.repository_label = QLabel()
        self.repository_label.setAlignment(Qt.AlignRight)
        self.widget.layout().addWidget(self.metaconfig_label, 0, 0)
        self.widget.layout().addWidget(self.repository_label, 0, 2)

    def paint(self, painter, option, index):
        option.index = index
        super().paint(painter, option, index)

    def drawDisplay(self, painter, option, rect, text):
        repository = option.index.data(int(IliDataItemModel.Roles.ILIREPO))
        models = option.index.data(int(IliDataItemModel.Roles.MODELS))
        owner = option.index.data(int(IliDataItemModel.Roles.OWNER))
        display_text = option.index.data(int(Qt.DisplayRole))

        self.repository_label.setText(
            '<font color="#666666"><i>of {owner} with {models} at {repository}</i></font>'.format(
                owner=owner, models="/".join(models), repository=repository
            )
        )
        self.metaconfig_label.setText(
            "{display_text}".format(display_text=display_text)
        )
        self.widget.setMinimumSize(rect.size())

        model_palette = option.palette
        if option.state & QStyle.State_Selected:
            model_palette.setColor(
                QPalette.WindowText,
                model_palette.color(QPalette.Active, QPalette.HighlightedText),
            )

        self.widget.render(painter, rect.topLeft(), QRegion(), QWidget.DrawChildren)


class IliToppingFileCache(IliDataCache):

    download_finished = pyqtSignal()
    """
    meta_netloc is the repository (netloc) of the metaconfiguration file used for file paths in the file_ids
    file_ids can contain ilidata: or file: information
    """

    def __init__(self, configuration, file_ids=None, tool_dir=None):
        IliDataCache.__init__(self, configuration)
        self.cache_path = os.path.expanduser("~/.ilitoppingfilescache")
        self.model = IliToppingFileItemModel()
        self.file_ids = file_ids
        self.tool_dir = (
            tool_dir
            if tool_dir
            else os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        )
        self.downloaded_files = list()
        self.file_download_succeeded.connect(
            lambda dataset_id, path: self.on_download_status(dataset_id)
        )
        self.file_download_failed.connect(
            lambda dataset_id, path: self.on_download_status(dataset_id)
        )
        self.model.rowsInserted.connect(lambda: self.on_download_status(None))

    def refresh(self):
        if not self.directories is None:
            for directory in self.directories:
                self.process_model_directory(directory)

        # collect local files
        netloc = "local_files"
        repo_files = list()
        for file_path_id in [
            file_id for file_id in self.file_ids if file_id[0:5] == "file:"
        ]:
            toppingfile = dict()
            toppingfile["id"] = file_path_id
            toppingfile["version"] = None
            toppingfile["owner"] = None
            toppingfile["repository"] = netloc
            toppingfile["url"] = None
            toppingfile["relative_file_path"] = file_path_id[5:]
            toppingfile["local_file_path"] = (
                file_path_id[5:]
                if os.path.isabs(file_path_id[5:])
                else os.path.join(self.tool_dir, file_path_id[5:])
            )
            if os.path.exists(toppingfile["local_file_path"]):
                self.file_download_succeeded.emit(
                    file_path_id, toppingfile["local_file_path"]
                )
            else:
                self.file_download_failed.emit(
                    file_path_id,
                    self.tr("Could not find local file  {}").format(file_path_id[5:]),
                )
            repo_files.append(toppingfile)

        self.repositories[netloc] = repo_files
        self.model.set_repositories(self.repositories)

    def on_download_status(self, dataset_id):
        # here we could add some more logic
        if dataset_id is not None:
            self.downloaded_files.append(dataset_id)
        if len(self.downloaded_files) == len(self.file_ids) == self.model.rowCount():
            self.download_finished.emit()

    def _process_informationfile(self, file, netloc, url):
        """
        Parses ilidata.xml provided in ``file`` and updates the local repositories cache.
        """
        try:
            root = ET.parse(file).getroot()
        except ET.ParseError as e:
            QgsMessageLog.logMessage(
                self.tr(
                    "Could not parse ilidata file `{file}` ({exception})".format(
                        file=file, exception=str(e)
                    )
                ),
                self.tr("QGIS Model Baker"),
            )
            return

        self.repositories[netloc] = list()
        repo_files = list()
        for repo in root.iter(
            "{http://www.interlis.ch/INTERLIS2.3}DatasetIdx16.DataIndex"
        ):
            for topping_metadata in repo.findall(
                "ili23:DatasetIdx16.DataIndex.DatasetMetadata", self.ns
            ):
                dataset_id = "ilidata:{}".format(
                    self.get_element_text(topping_metadata.find("ili23:id", self.ns))
                )
                if dataset_id in self.file_ids:
                    for files_element in topping_metadata.findall(
                        "ili23:files", self.ns
                    ):
                        for data_file in files_element.findall(
                            "ili23:DatasetIdx16.DataFile", self.ns
                        ):
                            for file_element in data_file.findall(
                                "ili23:file", self.ns
                            ):
                                for file in file_element.findall(
                                    "ili23:DatasetIdx16.File", self.ns
                                ):
                                    path = self.get_element_text(
                                        file.find("ili23:path", self.ns)
                                    )

                                    toppingfile = dict()
                                    toppingfile["id"] = dataset_id

                                    toppingfile["version"] = self.get_element_text(
                                        topping_metadata.find("ili23:version", self.ns)
                                    )
                                    toppingfile["owner"] = self.get_element_text(
                                        topping_metadata.find("ili23:owner", self.ns)
                                    )

                                    toppingfile["repository"] = netloc
                                    # relative_file_path like layerstyle/something.qml
                                    toppingfile["relative_file_path"] = path
                                    # url like http://models.opengis.ch or /home/nyuki/folder
                                    toppingfile["url"] = url
                                    toppingfile["local_file_path"] = self.download_file(
                                        netloc, url, path, dataset_id
                                    )
                                    repo_files.append(toppingfile)

        self.repositories[netloc] = sorted(
            repo_files,
            key=lambda m: m["version"] if m["version"] else "0",
            reverse=True,
        )

        self.model.set_repositories(self.repositories)


class IliToppingFileItemModel(QStandardItemModel):
    class Roles(Enum):
        ILIREPO = Qt.UserRole + 1
        VERSION = Qt.UserRole + 2
        RELATIVEFILEPATH = Qt.UserRole + 3
        LOCALFILEPATH = Qt.UserRole + 4
        OWNER = Qt.UserRole + 5
        URL = Qt.UserRole + 6

        def __int__(self):
            return self.value

    def __init__(self, parent=None):
        super().__init__(0, 1, parent)

    def set_repositories(self, repositories):
        self.clear()
        row = 0
        ids = list()

        for repository in repositories.values():

            for toppingfile in repository:
                if any(toppingfile["id"] == s for s in ids):
                    continue
                item = QStandardItem()
                item.setData(toppingfile["id"], int(Qt.DisplayRole))
                item.setData(toppingfile["id"], int(Qt.EditRole))
                item.setData(
                    toppingfile["repository"],
                    int(IliToppingFileItemModel.Roles.ILIREPO),
                )
                item.setData(
                    toppingfile["version"], int(IliToppingFileItemModel.Roles.VERSION)
                )
                item.setData(toppingfile["url"], int(IliToppingFileItemModel.Roles.URL))
                item.setData(
                    toppingfile["relative_file_path"],
                    int(IliToppingFileItemModel.Roles.RELATIVEFILEPATH),
                )
                item.setData(
                    toppingfile["local_file_path"],
                    int(IliToppingFileItemModel.Roles.LOCALFILEPATH),
                )
                item.setData(
                    toppingfile["owner"], int(IliToppingFileItemModel.Roles.OWNER)
                )

                ids.append(toppingfile["id"])
                self.appendRow(item)
                row += 1



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 2022-07-17
        git sha              : :%H$
        copyright            : (C) 2022 by Dave Signer
        email                : david at opengis ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import datetime
import os

from ..libs.toppingmaker import Target
from ..libs.toppingmaker.utils import slugify


class IliTarget(Target):
    """
    Extended class of toppingmaker.Target containing additional parameters like owner, publishing_date and version.
    And a path_resolver adding an id to the toppingfile in the toppingfile_list.
    """

    def __init__(
        self,
        projectname: str = "project",
        main_dir: str = None,
        sub_dir: str = None,
        path_resolver=None,
        owner="owner",
        publishing_date=None,
        version=None,
    ):
        if not path_resolver:
            path_resolver = self.ilidata_path_resolver
        super().__init__(projectname, main_dir, sub_dir, path_resolver)
        self.default_owner = owner
        self.default_publishing_date = (
            publishing_date or datetime.datetime.now().strftime("%Y-%m-%d")
        )
        self.default_version = version or datetime.datetime.now().strftime("%Y-%m-%d")

    @staticmethod
    def ilidata_path_resolver(target, name, type):
        """
        A path_resolver adding an id to the toppingfile in the toppingfile_list.

        The id is created with the type (like defintionfile, metaconfig etc.) and the filename.
        For uniqueness there is an incrementing number appended.

        Returns the id with the prefix "ilidata:". When using it as link, modelbaker knows that it has to look the id up in the ilidata.xml.
        """
        _, relative_filedir_path = target.filedir_path(type)

        id = target.unique_id_in_target_scope(target, slugify(f"{type}_{name}_001"))
        path = os.path.join(relative_filedir_path, name)
        type = type
        toppingfile = {"id": id, "path": path, "type": type}
        target.toppingfileinfo_list.append(toppingfile)
        return f"ilidata:{id}"

    @staticmethod
    def unique_id_in_target_scope(target, id):
        for toppingfileinfo in target.toppingfileinfo_list:
            if "id" in toppingfileinfo and toppingfileinfo["id"] == id:
                iterator = int(id[-3:])
                iterator += 1
                id = f"{id[:-3]}{iterator:03}"
                return target.unique_id_in_target_scope(target, id)
        return id



IGNORED_FIELDNAMES = ["t_id", "t_seq", "t_ili_tid", "T_Id", "T_Seq", "T_Ili_Tid"]

BASKET_FIELDNAMES = ["t_basket", "T_basket"]

READONLY_FIELDNAMES = ["t_ili_tid"]

# Some GeoPackage clients might add tables that we need to ignore based on
# prefix-suffix matches, here we define some known prefix-suffix pairs
GPKG_FILTER_TABLES_MATCHING_PREFIX_SUFFIX = [
    {
        "prefix": "rtree_",
        "suffix": [
            "_geometry",
            "_geometry_node",
            "_geometry_parent",
            "_geometry_rowid",
            "_geom",
            "_geom_node",
            "_geom_parent",
            "_geom_rowid",
        ],
    },
    {"prefix": "vgpkg_", "suffix": []},
]



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    13/05/19
    git sha              :    :%H$
    copyright            :    (C) 2019 by Yesid Polania
    email                :    yesidpol.3@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from abc import ABC, abstractmethod

from ..iliwrapper.ili2dbconfig import (
    Ili2DbCommandConfiguration,
    SchemaImportConfiguration,
)


class DbCommandConfigManager(ABC):
    """Manages a configuration object to return specific information of some database. This is a abstract class.

    Provides database uri, arguments to ili2db and a way to save and load configurations parameters
    based on a object configuration.

    :ivar configuration object that will be managed
    """

    def __init__(self, configuration: Ili2DbCommandConfiguration):
        """
        :param configuration: Configuration object that will be managed.
        :type configuration: :class:`Ili2DbCommandConfiguration`
        """
        self.configuration = configuration

    @abstractmethod
    def get_uri(self, su: bool = False, qgis: bool = False):
        """Gets database uri (connection string) for db connectors (:class:`DBConnector`).

        :param bool su: *True* to use super user credentials, *False* otherwise.
        :param bool qgis: *True* to use qgis specific credentials (e.g. authcfg), *False* otherwise.
        :return: Database uri (connection string).
        :rtype str
        """

    @abstractmethod
    def get_db_args(self, hide_password=False, su=False):
        """Gets a list of ili2db arguments related to database.

        :param bool hide_password: *True* to mask the password, *False* otherwise.
        :param bool su: *True* to use super user password, *False* otherwise. Default is False.
        :return: ili2db arguments list.
        :rtype: list
        """

    def get_schema_import_args(self):
        """Gets a list of ili2db arguments to use in operation schema import.

        :return: ili2db arguments list.
        :rtype: list
        """
        return list()

    def get_ili2db_args(self, hide_password=False):
        """Gets a complete list of ili2db arguments in order to execute the app.

        :param bool hide_password: *True* to mask the password, *False* otherwise.
        :return: ili2db arguments list.
        :rtype: list
        """
        db_args = self.get_db_args(hide_password, self.configuration.db_use_super_login)

        if type(self.configuration) is SchemaImportConfiguration:
            db_args += self.get_schema_import_args()

        ili2dbargs = self.configuration.to_ili2db_args(db_args)

        return ili2dbargs

    @abstractmethod
    def save_config_in_qsettings(self):
        """Saves configuration values related to database in QSettings.

        :return: None
        """

    @abstractmethod
    def load_config_from_qsettings(self):
        """Loads configuration values related to database from Qsettings.

        :return: None
        """



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    09/05/19
    git sha              :    :%H$
    copyright            :    (C) 2019 by Yesid Polania (BSF Swissphoto)
    email                :    yesidpol.3@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from .layer_uri import LayerUri


class MssqlLayerUri(LayerUri):
    def __init__(self, uri):
        LayerUri.__init__(self, uri)
        self.provider = "mssql"

    def get_data_source_uri(self, record):
        if record["geometry_column"]:
            data_source_uri = '{uri} key={primary_key} estimatedmetadata=true srid={srid} type={type} table="{schema}"."{table}" ({geometry_column}) sql='.format(
                uri=self._get_layer_uri_common(),
                primary_key=record["primary_key"],
                srid=record["srid"],
                type=record["type"],
                schema=record["schemaname"],
                table=record["tablename"],
                geometry_column=record["geometry_column"],
            )
        else:
            data_source_uri = '{uri} key={primary_key} estimatedmetadata=true srid=0 table="{schema}"."{table}" sql='.format(
                uri=self._get_layer_uri_common(),
                primary_key=record["primary_key"],
                schema=record["schemaname"],
                table=record["tablename"],
            )

        return data_source_uri

    def _get_layer_uri_common(self):
        param_db = dict()
        lst_item = self.uri.split(";")
        for item in lst_item:
            key_value = item.split("=")
            if len(key_value) == 2:
                key = key_value[0].strip()
                value = key_value[1].strip()
                param_db[key] = value

        uri = "service='driver={drv}' dbname='{database}' host={server} user='{uid}' password='{pwd}' ".format(
            drv=param_db["DRIVER"],
            database=param_db["DATABASE"],
            server=param_db["SERVER"],
            uid=param_db["UID"],
            pwd=param_db["PWD"],
        )

        return uri
